'''
Octree space curve simplification
=================================

Module for simplifying lines with an octree decomposition.
'''

import numpy as n
try:
    from coctree import (angle_exceeds as cangle_exceeds,
                         line_to_segments as cline_to_segments)
except ImportError:
    cangle_exceeds = None
    cline_to_segments = None


class OctreeError(Exception):
    pass


class OctreeCell(object):
    '''Stores line segments, and is capable of splitting and simplifying
    them.

    - lines must be a list of LineSegments. If your data is in another form,
      consider using the from_line or from_cell_lines class methods.
    - shape is the size of the cell
    - depth is the current recursion depth
    - state is a dictionary that stores information from the recursion tree
    - min_cell is the minimum cell side length
    - max_depth is the maximum recursion depth beyond which simplification is
      not attempted
    - top_level should be True if this OctreeCell is the first created, in which
      case a LineHandle is instantiated for each segment so that it can be
      reconstructed after simplification.
    - handles is a list of Handles to use, if it doesn't exist and top_level
      is True then one is created per segment. This *will* fail with some
      input, so use the class methods for building the cell to automatically
      pass the right handles.
    - cut_selection is a string saying how to select the cut planes in self.
      It may be 'regular' (8 equally sized octants), 'uniform' (values
      selected from a uniform distribution in each axis), or 'com' (values
      selected to pass through the line centre of mass of the cell).
    - is_knotted_func is an optional function to test for knotting.

    The Handle system is used to ensure correct behaviour (with simple code!)
    even for non-continuous segments etc.
    '''
    def __init__(self, lines, shape, depth=0, top_level=True,
                 state=None, min_cell=1., max_depth=100,
                 handles=None, cut_selection='regular',
                 is_knotted_func=None):

        self.depth = depth
        self.shape = shape
        self.max_depth = max_depth
        self.min_cell = min_cell

        self.cut_selection = cut_selection
        if cut_selection == 'regular':
            self.cut_planes = self.get_equidistant_planes()
        elif cut_selection == 'uniform':
            self.cut_planes = self.get_uniform_random_planes()
        elif cut_selection == 'com':
            raise OctreeError('com cut selection not yet supported')
        else:
            raise OctreeError('invalid choice of cut_selection')

        if state is None:
            state = {}
        self.state = state

        self.is_knotted_func = is_knotted_func

        if 'boundaries' not in state:
            state['boundaries'] = []
        state['boundaries'].append(self.boundary_lines())

        self.original_segments = lines
        self.segments = lines

        if handles is None:
            handles = []
        self.handles = handles
        if top_level and not self.handles:
            # Make a Handle to keep track of each segment even when
            # it is replaced with multiple shorter segments
            self.handles = [Handle(segment, segment.prev) for
                            segment in self.segments]
            for handle in self.handles:
                handle.next.prev = handle
                handle.prev.next = handle

    @classmethod
    def from_single_line(cls, line, shape=None, **kwargs):
        '''Returns an OctreeCell created from the given line, optionally
        with the given boundingbox shape. If the shape is not provided,
        it is calculated from the line extents.

        Other kwargs are passed directly to the OctreeCell.
        '''
        if shape is None:
            shape = [n.min(line[:, 0]), n.max(line[:, 0]),
                     n.min(line[:, 1]), n.max(line[:, 1]),
                     n.min(line[:, 2]), n.max(line[:, 2])]
        lineseg = LineSegment(line, identifier=1)
        lineseg.next = lineseg
        lineseg.prev = lineseg

        return cls([lineseg], shape, **kwargs)


    @classmethod
    def from_lines(cls, lines, shape=None, **kwargs):
        '''Returns an OctreeCell created from the given lines, optionally
        with the given boundingbox shape. If the shape is not provided,
        it is calculated from the maximum line extents.

        Other kwargs are passed directly to the OctreeCell.
        '''
        if shape is None:
            extremes = [[n.min(line[:, 0]), n.max(line[:, 0]),
                         n.min(line[:, 1]), n.max(line[:, 1]),
                         n.min(line[:, 2]), n.max(line[:, 2])] for
                         line in lines]
            extremes = n.array(extremes)
            shape = [n.min(extremes[:, 0]), n.max(extremes[:, 1]),
                     n.min(extremes[:, 2]), n.max(extremes[:, 3]),
                     n.min(extremes[:, 4]), n.max(extremes[:, 5])]
        segments = []
        for identifier, line in enumerate(lines):
            lineseg = LineSegment(line, identifier=identifier)
            lineseg.next = lineseg
            lineseg.prev = lineseg
            segments.append(lineseg)

        # import ipdb
        # ipdb.set_trace()

        return cls(segments, shape, **kwargs)
            


    @classmethod
    def from_cell_lines(cls, lines, shape, **kwargs):
        '''Returns an OctreeCell created from the given list of lines,
        which are assumed to be periodic.

        The shape may be all 6 values (xmin, xmax, ymin, ymax, zmin, zmax)
        or just xmax, ymax, zmax with the assumption that the minima are
        all zero.
        '''

        if len(shape) == 3:
            shape = (0, shape[0], 0, shape[1], 0, shape[2])
        short_shape = shape[1::2]
        
        if isinstance(lines[0], list):
            line_segs = lines
        else:
            line_segs = [
                split_cell_line(line, [s - 10. for s in short_shape])
                for line in lines]
        all_joined_segments = []
        handles = []
        for identifier, segments in enumerate(line_segs):
            segments = filter(lambda j: len(j) > 1, segments)
            seg_classes = [LineSegment(seg, identifier=identifier)
                           for seg in segments]
            handle = Handle(seg_classes[0])
            for i, seg_class in enumerate(seg_classes):
                nex = seg_classes[(i+1) % len(seg_classes)]
                # The CellJumpSegments isn't in self.segments, so it
                # won't be simplified
                jump_segment = CellJumpSegment(  
                    n.vstack((seg_class.points[-1], nex.points[0])))
                seg_class.next = jump_segment
                nex.prev = jump_segment
                jump_segment.prev = seg_class
                jump_segment.next = nex
            handle.prev = handle.next.prev
            handle.next.prev = handle
            handle.prev.next = handle
            handles.append(handle)

            all_joined_segments.extend(seg_classes)

        return cls(all_joined_segments, shape, handles=handles, **kwargs)

    def simplify(self, obey_knotting=True):
        '''
        Performs a suitable action depending on the segments in the cell:
        - if 1 segment, checks the total angle and reduces to a straight
          line if it is less than 3.8pi (4pi is needed for a knot)

        - otherwise, does the octant simplification if the cell min size
          and recursion depth have both not been hit

        The `obey_knotting` parameter may be False to ignore self-entangling
        of lines, i.e. a single line may just pass through itself so no
        angle check is necessary.
        '''

        segments = self.segments
        do_octants = False
        shape = self.shape
        size = self.get_cell_size()

        # Do something reasonable if there are two segments that include
        # the line start point
        if len(segments) == 2:
            s1, s2 = segments
            if s1.next == s2.prev:
                if isinstance(s1.next, Handle):
                    handle = s1.next
                    points = n.vstack((s1.points, s2.points))
                    new_segment = LineSegment(points,
                                              identifier=s1.identifier)
                    handle.prev = s1.prev
                    handle.prev.next = handle
                    handle.next = new_segment
                    new_segment.prev = handle
                    new_segment.next = s2.next
                    new_segment.next.prev = new_segment
                    self.segments = segments = [new_segment]
            elif s2.next == s1.prev:
                if isinstance(s2.next, Handle):
                    handle = s2.next
                    points = n.vstack((s2.points, s1.points))
                    new_segment = LineSegment(points,
                                              identifier=s2.identifier)
                    handle.prev = s2.prev
                    handle.prev.next = handle
                    handle.next = new_segment
                    new_segment.prev = handle
                    new_segment.next = s1.next
                    new_segment.next.prev = new_segment
                    self.segments = segments = [new_segment]

        if (len(segments) == 0 or self.depth >= self.max_depth or
            any([span < self.min_cell for span in size])):
            # If any of these are True, no simplification is possible
            return

        if not obey_knotting:
            identifiers = [segment.identifier for segment in segments]
            if None not in identifiers and len(set(identifiers)) == 1:
                for segment in segments:
                    segment.replace_with_straight_line()
                return


        elif len(segments) == 1:
            segment = segments[0]
            if not obey_knotting:
                segment.replace_with_straight_line()
                return
            if not angle_exceeds_func(segment.points, 2.*n.pi, False):
                segment.replace_with_straight_line()
                return
            if (self.is_knotted_func is not None and
                not self.is_knotted_func(segment.points)):
                segment.replace_with_straight_line()
                return

        # If we get here, the cell is too complex to simplify topologically,
        # but we are allowed to split into octants
                
        self._simplify_via_octants(obey_knotting)
        
    def _simplify_via_octants(self, obey_knotting=True):
        '''Splits each segment of self into sub-segments in each octant
        of self, and resolves them with a new OctreeCell for each octant.'''

        segments = self.segments
        cuts = self.cut_planes


        # Get new segments to be sorted into the new octants
        new_segments = []
        for seg in segments:
            new_segs = seg.cut_at(cuts)
            for each_new_seg in new_segs:
                new_segments.append(each_new_seg)

        # Place the segments in appropriate octants
        octant_contents = find_octants_of_segments(new_segments,
                                                   self.cut_planes)

        # Make and run a new OctreeCell for each octant
        fur, flr, fll, ful, bur, blr, bll, bul = octant_contents
        xmin, xmax, ymin, ymax, zmin, zmax = self.shape
        cut_x, cut_y, cut_z = self.cut_planes

        shapes = ((cut_x, xmax, cut_y, ymax, cut_z, zmax),
                  (cut_x, xmax, ymin, cut_y, cut_z, zmax),
                  (xmin, cut_x, ymin, cut_y, cut_z, zmax),
                  (xmin, cut_x, cut_y, ymax, cut_z, zmax),
                  (cut_x, xmax, cut_y, ymax, zmin, cut_z),
                  (cut_x, xmax, ymin, cut_y, zmin, cut_z),
                  (xmin, cut_x, ymin, cut_y, zmin, cut_z),
                  (xmin, cut_x, cut_y, ymax, zmin, cut_z))

        for i in range(8):
            octant_segments = octant_contents[i]
            if not len(octant_segments):
                continue
            shape = shapes[i]
            cxmin, cxmax, cymin, cymax, czmin, czmax = shape
            cell_size = (cxmax-cxmin) + (cymax-cymin) + (czmax-czmin)

            q = OctreeCell(octant_segments, shape, self.depth+1,
                           top_level=False,
                           state=self.state, min_cell=self.min_cell,
                           max_depth=self.max_depth,
                           cut_selection=self.cut_selection)
            q.simplify(obey_knotting)

    
    def get_equidistant_planes(self):
        '''Returns the x, y and z values halfway along each axis of the
        cell, according to self.shape.'''
        xmin, xmax, ymin, ymax, zmin, zmax = self.shape
        cut_x = (xmax + xmin) / 2.
        cut_y = (ymax + ymin) / 2.
        cut_z = (zmax + zmin) / 2.
        return (cut_x, cut_y, cut_z)

    def get_uniform_random_planes(self):

        '''Returns x, y and z values each uniformly randomly distributed
        through self.shape.'''
        xmin, xmax, ymin, ymax, zmin, zmax = self.shape
        return (n.array([xmin, ymin, zmin]) +
                (0.1 + 0.8*n.random.random(3)) *
                n.array(self.get_cell_size()))

    def get_cell_size(self):
        return (self.shape[1] - self.shape[0],
                self.shape[3] - self.shape[2],
                self.shape[5] - self.shape[4])
        
    def boundary_lines(self):
        '''Get a line tracing the boundary of the cell. These lines
        are passed recursively up the OctreeCell tree so that the top level one
        keeps a list of all of them - useful for plotting the tree state
        later.'''
                 
        xmin, xmax, ymin, ymax, zmin, zmax = self.shape
        return n.array([[xmin, ymin, zmin],
                        [xmax, ymin, zmin],
                        [xmax, ymax, zmin],
                        [xmin, ymax, zmin],
                        [xmin, ymin, zmin],
                        [xmin, ymin, zmax],
                        [xmax, ymin, zmax],
                        [xmax, ymin, zmin],
                        [xmax, ymin, zmax],
                        [xmax, ymax, zmax],
                        [xmax, ymax, zmin],
                        [xmax, ymax, zmax],
                        [xmin, ymax, zmax],
                        [xmin, ymax, zmin],
                        [xmin, ymax, zmax],
                        [xmin, ymin, zmax]])

    def get_lines(self, remove_unnecessary_jumps=True):
        '''Return all the lines of self by joining any segments back
        together again.'''
        handles = self.handles
        lines = [handle.reconstruct_line(remove_unnecessary_jumps)
                 for handle in handles]
        return lines

    def get_single_line(self, remove_unnecessary_jumps=True):
        '''Return the first line in the set of lines reconstructed from
        segments in self. This is useful for OctreeCells that only had
        a single line to begin with, e.g. simplifying a knot.
        '''
        lines = self.get_lines(remove_unnecessary_jumps)
        if len(lines) == 0:
            raise OctreeError('Tried to get single line but no lines found.')
        return lines[0]


        
class Handle(object):
    '''Simply stores a next and previous LineSegment, with a method to
    reconstruct the line by walking along the segments.

    The Handle is like a LineSegment with no point data - all it does is
    sit in the line keeping track of its next and prev segments (which
    may be modified by the cut_at method of those LineSegments). Since
    it will never change itself, it can be used later to rejoin all these
    LineSegments which may have been created arbitrarily deep in the
    decomposition, returning the same line that corresponds to its input!
    '''

    def __init__(self, next=None, prev=None, identifier=None):
        self.next = next  
        self.prev = prev  
        self.identifier = identifier

    def reconstruct_line(self, remove_unnecessary_jumps=True):
        '''Returns the line stored by the handle, by walking along it until
        returning to the Handle or reaching a segment with no next entry.

        If ``remove_unnecessary_jumps`` is True, will remove jumps connected
        by a trivial segment along the periodic cell wall.
        '''
        comp_segs = [self.next]
        current = self.next
        while current.next is not self and current.next is not None:
            current = current.next
            comp_segs.append(current)

        if remove_unnecessary_jumps:
            invalid_indices = []
            for i, segment in enumerate(comp_segs):
                if i not in invalid_indices:
                    segment2 = comp_segs[(i+1) % len(comp_segs)]
                    segment3 = comp_segs[(i+1) % len(comp_segs)]
                    if (isinstance(segment, CellJumpSegment) and
                        isinstance(segment3, CellJumpSegment) and
                        len(segment2.points) == 2):
                        invalid_indices.append(i)

            comp_segs = [seg for i, seg in enumerate(comp_segs) if
                         i not in invalid_indices]

        return resample(n.vstack([seg.points[:-1] for
                                  seg in comp_segs] +
                                 [comp_segs[0].points[:1] +
                                  0.00001]))

    def get_line_components(self):
        '''Returns a list of all the LineSegments in the line with
        this handle.'''
        component_segments = [self.next]
        current = self.next
        while current.next is not self and current.next is not None:
            current = current.next
            component_segments.append(current)
        return component_segments

        

class LineSegment(object):
    '''Stores a section of line, and can split at boundaries.'''

    def __init__(self, points, next=None, prev=None, identifier=None):
        self.points = points
        self.next = next  # the next LineSegment
        self.prev = prev  # the previous LineSegment

        self.identifier = identifier  # identifies all segments of the
                                      # same line


    def cell(self, cut_planes):
        '''Returns a string representing where the LineSegment is,
        in terms of which octant of the space divided by cut_planes
        x, y, z the LineSegment appears in.

        This assumes the LineSegment is only within one octant (except
        perhaps points on the boundary).
        '''

        cut_x, cut_y, cut_z = cut_planes
        p = self.points[0] + 0.51123*(self.points[1] - self.points[0])
        x, y, z = p
        if z > cut_z:
            if x > cut_x and y > cut_y:
                return 'fur'
            elif x > cut_x and y <= cut_y:
                return 'flr'
            elif x <= cut_x and y <= cut_y:
                return 'fll'
            elif x <= cut_x and y > cut_y:
                return 'ful'
            else:
                raise OctreeError('Segment not in any region?')
        elif z <= cut_z:
            if x > cut_x and y >= cut_y:
                return 'bur'
            elif x > cut_x and y <= cut_y:
                return 'blr'
            elif x <= cut_x and y <= cut_y:
                return 'bll'
            elif x <= cut_x and y > cut_y:
                return 'bul'
            else:
                raise OctreeError('Segment not in any region?')
        else:
            raise OctreeError('Segment not in any region?')

    def cut_at(self, cuts):
        '''Returns a list of new LineSegments, resulting from cutting self at
        the given cut planes.
        '''
        points = self.points.copy()
        crude_segs = line_to_segments(points, cuts, join_ends=False)
        segs = crude_segs_to_segs(crude_segs, self.identifier)

        # crude_segs_to_segs assumes the segments form a loop
        # here we instead connect the first to self.prev, and
        # the last to self.next
        segs[0].prev = self.prev
        self.prev.next = segs[0]
        segs[-1].next = self.next
        self.next.prev = segs[-1]

        return segs

    def replace_with_straight_line(self):
        '''
        Replaces self.points with a straight line (i.e. just the first
        and last vertices).

        This is only called when only one segment lies in a quadcell,
        therefore components can be cut out safely without worrying
        about any other parts of the full curve.'''

        if self.next is not self.prev:
            self.points = n.vstack((self.points[0], self.points[-1]))
        else:  # if the line is a loop, just cut out most points
            self.points = self.points[::int(len(self.points)/3.)]


class CellJumpSegment(LineSegment):
    '''A special LineSegment that denotes a jump through periodic boundary
    conditions. This has no special behaviour of its own, but is
    subclassed in case this is useful later.
    '''


def line_to_segments(line, cuts=None, join_ends=True):
    '''Takes a line (set of points), a list of cut planes in
    x, y, z, and a parameter to decide whether the line
    joining the first and last point should also be cut.

    Returns a list of shorter lines resulting from cutting at
    all these cut planes.'''

    line = line.copy()

    if cuts is None:
        xmin = n.min(line[:,0]) - 1
        xmax = n.max(line[:,0]) + 1
        ymin = n.min(line[:,1]) - 1
        ymax = n.max(line[:,1]) + 1
        zmin = n.min(line[:,2]) - 1
        zmax = n.max(line[:,2]) + 1

        cut_x = (xmax + xmin) / 2.
        cut_y = (ymax + ymin) / 2.
        cut_z = (zmin + zmax) / 2.
    else:
        cut_x, cut_y, cut_z = cuts

    # Cut the line wherever it passes through a quad cell boundary
    segments = []
    cut_i = 0
    for i in range(len(line)-1):
        cur = line[i]
        nex = line[i+1]

        dv = nex - cur
        dx, dy, dz = dv

        cross_cut_x = n.sign(cur[0] - cut_x) != n.sign(nex[0] - cut_x)
        cross_cut_y = n.sign(cur[1] - cut_y) != n.sign(nex[1] - cut_y)
        cross_cut_z = n.sign(cur[2] - cut_z) != n.sign(nex[2] - cut_z)

        if cross_cut_x and cross_cut_y and cross_cut_z:
            x_cut_pos = -1 * (cur[0]-cut_x)/dx
            y_cut_pos = -1 * (cur[1]-cut_y)/dy
            z_cut_pos = -1 * (cur[2]-cut_z)/dz
            order = n.sort((x_cut_pos, y_cut_pos, z_cut_pos))
            # assert 0 < x_cut_pos < 1 and 0 < y_cut_pos < 1 and 0 < z_cut_pos < 1
            join_point_1 = cur + order[0]*dv
            join_point_2 = cur + order[1]*dv
            join_point_3 = cur + order[2]*dv
            first_seg = n.vstack((line[cut_i:(i+1)].copy(), join_point_1))
            second_seg = n.vstack((join_point_1, join_point_2))
            third_seg = n.vstack((join_point_2, join_point_3))
            line[i] = join_point_3
            cut_i = i
            segments.append(first_seg)
            segments.append(second_seg)
            segments.append(third_seg)
        elif cross_cut_x and cross_cut_y:
            x_cut_pos = -1 * (cur[0]-cut_x)/dx
            y_cut_pos = -1 * (cur[1]-cut_y)/dy
            order = n.sort((x_cut_pos, y_cut_pos))
            join_point_1 = cur + order[0]*dv
            join_point_2 = cur + order[1]*dv
            first_seg = n.vstack((line[cut_i:(i+1)].copy(), join_point_1))
            second_seg = n.vstack((join_point_1, join_point_2))
            line[i] = join_point_2
            cut_i = i
            segments.append(first_seg)
            segments.append(second_seg)
        elif cross_cut_x and cross_cut_z:
            x_cut_pos = -1 * (cur[0]-cut_x)/dx
            z_cut_pos = -1 * (cur[2]-cut_z)/dz
            order = n.sort((x_cut_pos, z_cut_pos))
            join_point_1 = cur + order[0]*dv
            join_point_2 = cur + order[1]*dv
            first_seg = n.vstack((line[cut_i:(i+1)].copy(), join_point_1))
            second_seg = n.vstack((join_point_1, join_point_2))
            line[i] = join_point_2
            cut_i = i
            segments.append(first_seg)
            segments.append(second_seg)
        elif cross_cut_y and cross_cut_z:
            y_cut_pos = -1 * (cur[1]-cut_y)/dy
            z_cut_pos = -1 * (cur[2]-cut_z)/dz
            order = n.sort((y_cut_pos, z_cut_pos))
            join_point_1 = cur + order[0]*dv
            join_point_2 = cur + order[1]*dv
            first_seg = n.vstack((line[cut_i:(i+1)].copy(), join_point_1))
            second_seg = n.vstack((join_point_1, join_point_2))
            line[i] = join_point_2
            cut_i = i
            segments.append(first_seg)
            segments.append(second_seg)
        elif cross_cut_x:
            cut_pos = -1 * (cur[0]-cut_x)/dx
            assert 0. <= cut_pos <= 1.
            join_point = cur + cut_pos*dv
            first_seg = n.vstack((line[cut_i:(i+1)].copy(), join_point))
            line[i] = join_point
            cut_i = i
            segments.append(first_seg)
        elif cross_cut_y:
            cut_pos = -1 * (cur[1]-cut_y)/dy
            assert 0. <= cut_pos <= 1.
            join_point = cur + cut_pos*dv
            first_seg = n.vstack((line[cut_i:(i+1)].copy(), join_point))
            line[i] = join_point
            cut_i = i
            segments.append(first_seg)
        elif cross_cut_z:
            cut_pos = -1 * (cur[2]-cut_z)/dz
            assert 0. <= cut_pos <= 1.
            join_point = cur + cut_pos*dv

            first_seg = n.vstack((line[cut_i:(i+1)].copy(), join_point))
            line[i] = join_point
            # second_seg = n.vstack((join_point, line[(i+1):]))

            cut_i = i
            segments.append(first_seg)

    final_seg = line[cut_i:]
    if cut_i > 0:
        if join_ends:
            first_seg = segments.pop(0)
            segments.append(n.vstack((final_seg, first_seg)))
        else:
            segments.append(final_seg)
    else:
        segments.append(final_seg)

    return segments


def crude_segs_to_segs(csegs, identifier=None):
    '''Takes a list of line segments and makes them into LineSegment
    instances, under the assumption that the last is joined to the
    first. The segments are labelled by the passed identifier.
    '''
    segs = []
    for s in csegs:
        s = LineSegment(s, identifier=identifier)
        segs.append(s)
    for i in range(len(segs)):
        cur = segs[i]
        prev = segs[(i-1) % len(segs)]
        next = segs[(i+1) % len(segs)]
        cur.prev = prev
        cur.next = next
    return segs


def angle_exceeds(ps, val=2*n.pi, include_closure=True):
    '''Returns True if the sum of angles along ps exceeds
    val, else False.

    If include_closure, includes the angles with the line closing
    the end and start points.
    '''
    angle = 0.
    nex = ps[0]
    nex2 = ps[1]
    dv2 = nex2-nex
    dv2 /= mag(dv2)
    checks = range(len(ps)) if include_closure else range(len(ps)-2)
    for i in checks:
        cur = nex
        nex = nex2
        nex2 = ps[(i+2) % len(ps)]
        dv = dv2
        dv2 = nex2-nex
        dv2 /= mag(dv2)
        increment = angle_between(dv, dv2)
        if n.isnan(increment):
            return True
        angle += increment
        if angle > val:
            return True
    assert not n.isnan(angle)
    return False

def angle_between(v1, v2):
    '''Returns angle between v1 and v2, assuming they are normalised to 1.'''
    # clip becaus v1.dot(v2) may exceed 1 due to floating point
    return n.arccos(n.clip(n.abs(v1.dot(v2)), 0., 1.))


def mag(v):
    return n.sqrt(v.dot(v))

def find_octants_of_segments(segments, cuts):
    '''For each LineSegment in the passed segments, returns an
    identifier for the octant it should be placed in'''
    fur = []  # front upper right etc.
    flr = []
    fll = []
    ful = []
    bur = []
    blr = []
    bll = []
    bul = []
    for seg in segments:
        cell = seg.cell(cuts)
        if cell == 'fur':
            fur.append(seg)
        elif cell == 'flr':
            flr.append(seg)
        elif cell == 'fll':
            fll.append(seg)
        elif cell == 'ful':
            ful.append(seg)
        elif cell == 'bur':
            bur.append(seg)
        elif cell == 'blr':
            blr.append(seg)
        elif cell == 'bll':
            bll.append(seg)
        elif cell == 'bul':
            bul.append(seg)
        else:
            raise OctreeError('Segment not in recognised cell.')
    return (fur, flr, fll, ful, bur, blr, bll, bul)

def resample(points):
    '''Takes a set of points, and cuts out the middle ones if they form
    part of a straight line, or if they are identical to the previous
    point.'''
    if len(points) < 2:
        return points

    changing = True
    while changing:
        changing = False
        length = len(points)
        points = points.copy()
        keep = n.ones(len(points), dtype=bool)
        cur = points[0]
        nex = points[1]
        dv2 = nex-cur
        dv2 /= mag(dv2)

        for i in range(1, len(points) - 1):
            prev = cur
            cur = nex
            nex = points[i+1]
            if n.all(n.less(n.abs(cur-nex), 0.0001)):
                keep[i] = False
            else:
                dv1 = cur - prev
                dv1 /= mag(dv1)
                dv2 = nex - cur
                dv2 /= mag(dv2)
                if n.abs(dv1.dot(dv2) - 1) < 0.0001:
                    keep[i] = False
        points = points[keep]
        new_length = len(points)
        if new_length != length:
            changing = True
    return points[keep]

def remove_nearby_points(points):
    '''Takes a set of points, and removes those that are no distance from
    the previous one.'''
    points = points.copy()
    keep = n.ones(len(points), dtype=n.bool)

    comparator = points[0]
    for i, point in enumerate(points):
        nex = points[(i+1) % len(points)]
        if n.all(n.abs((nex - point)) < 0.00001):
            keep[i] = False
    keep[-1] = True

    return points[keep]
                 

def split_cell_line(line, shape=(10, 10, 10.)):
    '''Takes a cell lines, and cuts at the periodic boundaries
    (in this represenation, this is where it jumps)'''
    shape = n.array(shape)
    line = line.copy()
    i = 0
    out = []
    while i < (len(line)-1):
        cur = line[i]
        nex = line[i+1]
        if n.any(n.abs(nex-cur) > shape):
            firsthalf = line[:(i+1)]
            secondhalf = line[(i+1):]
            out.append(firsthalf)
            line = secondhalf
            i = 0
        else:
            i += 1
    out.append(line)
    return out

# Setup functions that *may* depend on cython
angle_exceeds_func = angle_exceeds
# angle_exceeds_func = (cangle_exceeds if cangle_exceeds is not None
#                       else angle_exceeds)
line_to_segments_func = line_to_segments
# line_to_segments_func = (cline_to_segments if cline_to_segments is not None
#                       else line_to_segments)

def determinant_indicates_knot(points):
    if len(points) > 500:
        return True
    from pyknotid.spacecurves import Knot
    k = Knot(points)
    return (k.determinant() > 1)
        



'''
DTNotation
==========

Classes for working with DT notation representing planar projections
of curves.

API documentation
~~~~~~~~~~~~~~~~~
'''

import numpy as n
import re
import sys

if sys.version_info.major == 2:
    string_types = basestring
else:
    string_types = str

class DTNotation(object):
    '''
    Class for containing and manipulation DT notation.

    Parameters
    ----------
    code : str or array-like
        The DT code. Must be either a string of entries separated
        by spaces, or an array.
    '''

    def __init__(self, code):
        if isinstance(code, string_types):
            self._init_from_string(code)
        elif isinstance(code, n.ndarray):
            code = [code]
            self._dt = code
        elif isinstance(code, list):
            self._dt = code
        else:
            raise ValueError('Could not parse code as a valid type')

    def _init_from_string(self, code):
        '''
        Converts the string into internal DT notation representation.
        '''

        regex = re.compile('[\n]*')
        lines = regex.split(code)

        dt = []

        for line in lines:
            numbers = line.split(' ')
            dt.append(n.array([int(number) for number in numbers], dtype=n.int))

        self._dt = dt

    def gauss_code_string(self):
        '''Returns a string containing a Gauss code, in the format accepted
        by :class:`~pyknotid.representations.gausscode.GaussCode`.

        To get a :class:`~pyknotid.representations.gausscode.GaussCode`
        object, you can pass this string when initialising it, or use
        :meth:`DTNotation.representation`.

        '''
        if len(self._dt) > 1:
            raise ValueError('DTNotation -> GaussCode does not yet '
                             'work with links')

        dt = self._dt[0]
        arr = n.zeros((len(dt) * 2, 2), dtype=n.int)


        for index, even in enumerate(dt, 0):
            odd = 2*index
            sign = n.sign(even)
            even = n.abs(even) - 1

            arr[odd, 0] = index + 1
            arr[odd, 1] = sign

            arr[even, 0] = index + 1
            arr[even, 1] = -1 * sign

        str_entries = ['{}{}'.format(index, '+' if sign > 0 else '-')
                       for index, sign in arr]

        return ','.join(str_entries)

    def representation(self, **kwargs):
        '''Returns a
        :class:`~pyknotid.representations.representation.Representation`
        representing the same DT code. The crossing orientations (and
        therefore resulting chirality) are chosen arbitrarily.
        '''
        from pyknotid.representations import Representation
        return Representation.calculating_orientations(
            self.gauss_code_string(), **kwargs)

    # def space_curve(self):
    #     gc = 

    def __str__(self):
        return repr(self)

    def __repr__(self):
        if len(self._dt) == 0:
            return '----'
        strings = [' '.join(map(str, l)) for l in self._dt]
        return '\n'.join(strings)



'''Invariants
==========

Functions for retrieving invariants of knots and links.

Many of these functions can be called in a more convenient way via
methods of the :doc:`space curve classes <spacecurves/index>`
(e.g. :class:`~pyknotid.spacecurves.knot.Knot`) or the
:class:`~pyknotid.representations.representation.Representation` class.


Mathematica
-----------

Functions whose name ends with ``_mathematica`` try to create an
external Mathematica process to calculate the answer. They may hang
or have other problems if Mathematica isn't available in your
``$PATH``, so be careful using them.

.. warning:: This module may be broken into multiple components at
             some point.

API documentation
-----------------

'''
from __future__ import print_function
import subprocess
import re
import sympy as sym
import numpy as n

from pyknotid.utils import vprint

def alexander(representation, variable=-1, quadrant='lr', simplify=True,
              mode='python'):
    '''Calculates the Alexander polynomial of the given knot. The
    representation *must* have just one knot component, or the
    calculation will fail or potentially give bad results.

    The result is returned with whatever numerical precision the
    algorithm produces, it is not rounded.

    The given representation *must* be simplified (RM1 performed if
    possible) for this to work, otherwise the matrix has overlapping
    elements. This is so important that this function automatically
    calls
    :meth:`pyknotid.representations.gausscode.GaussCode.simplify`, you
    must disable this manually if you don't want to do it.

    .. note:: If 'maxima' or 'mathematica' is chosen as the mode, the
              variable will automatically be set to ``t``.

    .. note:: If the mode is 'cypari', the quadrant argument will be
              ignored and the upper-left quadrant always used.

    Parameters
    ----------
    representation : Anything convertible to a
                     :class:`~pyknotid.representations.gausscode.GaussCode`
        A pyknotid representation class for the knot, or anything that
        can automatically be converted into a GaussCode (i.e. by writing
        :code:`GaussCode(your_object)`).
    variable : float or complex or sympy variable
        The value to caltulate the Alexander polynomial at. Defaults to -1,
        but may be switched to the sympy variable ``t`` in the future.
        Supports int/float/complex types (fast, works for thousands of
        crossings) or sympy
        expressions (much slower, works mostly only for <100 crossings).
    quadrant : str
        Determines what principal minor of the Alexander matrix should be
        used in the calculation; all choices *should* give the same answer.
        Must be 'lr', 'ur', 'ul' or 'll' for lower-right, upper-right,
        upper-left or lower-left respectively.
    simplify : bool
        Whether to call the GaussCode simplify method, defaults to True.
    mode : string
        One of 'python', 'maxima', 'cypari' or 'mathematica'. 
        denotes what
        tools to use; if python, the calculation is performed with
        numpy or sympy as appropriate. If maxima or mathematica, that
        program is called by the function - this will only work if the
        external tool is installed and available. Defaults to python.

    '''

    from pyknotid.representations.gausscode import GaussCode
    if not isinstance(representation, GaussCode):
        representation = GaussCode(representation)

    if simplify:
        representation.simplify(one=True, two=True, one_extended=True)

    code_list = representation._gauss_code
    if len(code_list) == 0:
        return 1
    elif len(code_list) > 1:
        raise Exception('tried to calculate alexander polynomial'
                        'for something with more than 1 component')

    crossings = code_list[0]

    if len(crossings) == 0:
        return 1

    if quadrant not in ['lr', 'ur', 'ul', 'll']:
        raise Exception('invalid quadrant')

    mode = mode.lower()
    if mode == 'maxima':
        return alexander_maxima(representation, quadrant,
                                verbose=False,
                                simplify=False)
    elif mode == 'cypari':
        return alexander_cypari(representation, quadrant,
                                verbose=False,
                                simplify=False)
    elif mode == 'mathematica':
        return alexander_mathematica(representation, quadrant,
                                     verbose=False)
                                     

    if isinstance(variable, (int, float, complex)):
        return _alexander_numpy(crossings, variable, quadrant)
    else:
        return _alexander_sympy(crossings, variable, quadrant)


def _alexander_numpy(crossings, variable=-1.0, quadrant='lr'):
    '''
    Numpy implementation of the Alexander polynomial (evaluated
    at a float), assuming the input has been sanitised by
    :func:`alexander`.
    '''
    import numpy as n
    num_crossings = int(len(crossings)/2)
    dtype = n.complex if isinstance(variable, n.complex) else n.float
    matrix = n.zeros((num_crossings, num_crossings), dtype=dtype)
    line_num = 0
    crossing_num_counter = 0
    crossing_dict = {}
    crossing_exists = False

    over_clock = 1. - 1. / variable
    under_clock_after = 1. / variable
    over_aclock = 1 - variable
    under_aclock_after = variable

    for i in range(len(crossings)):
        identifier, over, clockwise = crossings[i]
        if identifier in crossing_dict:
            crossing_num = crossing_dict.pop(identifier)
            crossing_exists = True
        if not crossing_exists:
            crossing_num = crossing_num_counter
            crossing_num_counter += 1
            crossing_dict[identifier] = crossing_num
        crossing_exists = False

        if over > 0.99999:
            mat_elt = over_clock if clockwise > 0.999 else over_aclock
            matrix[crossing_num, line_num % num_crossings] = mat_elt
        else:
            new_mat_elt = (under_clock_after if clockwise > 0.999 else
                           under_aclock_after)
            matrix[crossing_num, line_num % num_crossings] = -1
            line_num += 1
            matrix[crossing_num, line_num % num_crossings] = new_mat_elt
    if quadrant == 'lr':
        poly_val = n.linalg.det(matrix[1:, 1:])
    elif quadrant == 'ur':
        poly_val = n.linalg.det(matrix[:-1, 1:])
    elif quadrant == 'ul':
        poly_val = n.linalg.det(matrix[:-1:, :-1])
    elif quadrant == 'll':
        poly_val = n.linalg.det(matrix[1:, :-1])
    if not isinstance(poly_val, n.complex):
        poly_val = n.abs(poly_val)
    return poly_val


def _alexander_sympy(crossings, variable=None, quadrant='lr'):
    '''
    Sympy implementation of the Alexander polynomial, evaluated
    with the variable replaced by some sympy expression.
    '''
    # This is almost the same as the numpy implementation...they should
    # probably be merged.
    import sympy as sym
    if variable is None:
        variable = sym.var('t')
    num_crossings = int(len(crossings)/2)
    matrix = sym.zeros(num_crossings)
    line_num = 0
    crossing_num_counter = 0
    crossing_dict = {}
    crossing_exists = False

    over_clock = 1 - 1 / variable
    under_clock_after = 1 / variable
    over_aclock = 1 - variable
    under_aclock_after = variable

    for i in range(len(crossings)):
        identifier, over, clockwise = crossings[i]
        if identifier in crossing_dict:
            crossing_num = crossing_dict.pop(identifier)
            crossing_exists = True
        if not crossing_exists:
            crossing_num = crossing_num_counter
            crossing_num_counter += 1
            crossing_dict[identifier] = crossing_num
        crossing_exists = False

        if over > 0.99999:
            mat_elt = over_clock if clockwise > 0.999 else over_aclock
            matrix[crossing_num, line_num % num_crossings] = mat_elt
        else:
            new_mat_elt = (under_clock_after if clockwise > 0.999 else
                           under_aclock_after)
            matrix[crossing_num, line_num % num_crossings] = -1
            line_num += 1
            matrix[crossing_num, line_num % num_crossings] = new_mat_elt
    if quadrant == 'lr':
        poly_val = matrix[1:, 1:].det()
    elif quadrant == 'ur':
        poly_val = matrix[:-1, 1:].det()
    elif quadrant == 'ul':
        poly_val = matrix[:-1:, :-1].det()
    elif quadrant == 'll':
        poly_val = matrix[1:, :-1].det()
    return poly_val


def alexander_maxima(representation, quadrant='ul', verbose=False,
                     simplify=True):
    '''
    Returns the Alexander polynomial of the given representation, by
    calculating the matrix determinant in maxima.

    The function only supports evaluating at the variable ``t``.

    Parameters
    ----------
    representation : Anything convertible to a
                     :class:`~pyknotid.representations.gausscode.GaussCode`
        A pyknotid representation class for the knot, or anything that
        can automatically be converted into a GaussCode (i.e. by writing
        :code:`GaussCode(your_object)`).
    quadrant : str
        Determines what principal minor of the Alexander matrix should be
        used in the calculation; all choices *should* give the same answer.
        Must be 'lr', 'ur', 'ul' or 'll' for lower-right, upper-right,
    verbose : bool
        Whether to print information about the procedure. Defaults to False.
    simplify : bool
        If True, tries to simplify the representation before calculating
        the polynomial. Defaults to True.
    '''

    from pyknotid.representations.gausscode import GaussCode
    if not isinstance(representation, GaussCode):
        representation = GaussCode(representation)

    if simplify:
        representation.simplify(one=True, two=True, one_extended=True)

    if len(representation._gauss_code) > 1:
        raise Exception('tried to calculate alexander polynomial'
                        'for something with more than 1 component')

    code = representation._gauss_code[0]

    if len(code) == 0:
        return 1

    mat_mat = _maxima_matrix(code, quadrant=quadrant, verbose=verbose)

    code = ''.join([#'sparse: true;\n',
                    'ratmx: true;\n',
                    'display2d: false;\n',
                    mat_mat,
                    'expand(determinant(m2));'])

    with open('maxima_batch.maxima', 'w') as fileh:
        fileh.write(code)

    result = subprocess.check_output(
        ['maxima', '-b', 'maxima_batch.maxima'])
    print('Maxima output is:\n', result)
    result = result.decode('utf-8').split('\n')[-3][6:]

    t = sym.var('t')

    return eval(result.replace('^', '**'))


def alexander_cypari(representation, quadrant='ul', verbose=False,
                     simplify=True):
    '''
    Returns the Alexander polynomial of the given representation, by
    calculating the matrix determinant via cypari, a python interface
    to Pari-GP.

    The function only supports evaluating at the variable ``t``.

    The returned object is a cypari query type.

    Parameters
    ----------
    representation : Anything convertible to a
                     :class:`~pyknotid.representations.gausscode.GaussCode`
        A pyknotid representation class for the knot, or anything that
        can automatically be converted into a GaussCode (i.e. by writing
        :code:`GaussCode(your_object)`).
    quadrant : str
        Determines what principal minor of the Alexander matrix should be
        used in the calculation; all choices *should* give the same answer.
        Must be 'lr', 'ur', 'ul' or 'll' for lower-right, upper-right,
    verbose : bool
        Whether to print information about the procedure. Defaults to False.
    simplify : bool
        If True, tries to simplify the representation before calculating
        the polynomial. Defaults to True.
    '''

    from pyknotid.representations.gausscode import GaussCode
    if not isinstance(representation, GaussCode):
        representation = GaussCode(representation)

    if simplify:
        representation.simplify(one=True, two=True, one_extended=True)

    if len(representation._gauss_code) > 1:
        raise Exception('tried to calculate alexander polynomial'
                        'for something with more than 1 component')

    code = representation._gauss_code[0]

    if len(code) == 0:
        return 1

    mat_mat = _cypari_matrix(code, quadrant=quadrant, verbose=verbose)

    return mat_mat.matdet()


                    
def alexander_mathematica(representation, quadrant='ul', verbose=False,
                          via_file=True):
    '''
    Returns the Alexander polynomial of the given representation, by
    creating a Mathematica process and running its knot routines.
    The Mathematica installation must include the KnotTheory package.

    The function only supports evaluating at the variable ``t``.

    Parameters
    ----------
    representation : Anything convertible to a
                     :class:`~pyknotid.representations.gausscode.GaussCode`
        A pyknotid representation class for the knot, or anything that
        can automatically be converted into a GaussCode (i.e. by writing
        :code:`GaussCode(your_object)`).
    quadrant : str
        Determines what principal minor of the Alexander matrix should be
        used in the calculation; all choices *should* give the same answer.
        Must be 'lr', 'ur', 'ul' or 'll' for lower-right, upper-right,
    verbose : bool
        Whether to print information about the procedure. Defaults to False.
    via_file : bool
        If True, calls Mathematica via a written file ``mathematicascript.m``,
        otherwise calls Mathematica directly with ``runMath``. The latter
        had a nasty bug in at least one recent Mathematica version, so the
        default is to True.
    simplify : bool
        If True, tries to simplify the representation before calculating
        the polynomial. Defaults to True.
    '''
    from pyknotid.representations.gausscode import GaussCode
    if not isinstance(representation, GaussCode):
        representation = GaussCode(representation)

    if simplify:
        representation.simplify(one=True, two=True, one_extended=True)

    code = representation._gauss_code

    if len(code) == 0:
        return 1
    elif len(code) > 1:
        raise Exception('tried to calculate alexander polynomial'
                        'for something with more than 1 component')
    mat_mat = _mathematica_matrix(code[0], quadrant=quadrant, verbose=verbose)
    if not via_file:
        det = subprocess.check_output(['runMath', 'Det[' + mat_mat + ']'])[:-1]
    else:
        _write_mathematica_script('mathematicascript.m', 'Print[Det[' +
                                  mat_mat + ']]')
        det = subprocess.check_output(['bash', 'mathematicascript.m'])
    t = sym.var('t')
    return eval(det.replace('^', '**'))


def jones_mathematica(representation):
    '''
    Returns the Jones polynomial of the given representation, by
    creating a Mathematica process and running its knot routines.
    The Mathematica installation must include the KnotTheory package.

    The function only supports evaluating at the variable ``q``.

    Parameters
    ----------
    representation : A PlanarDiagram, or anything convertible to a
                     :class:`~pyknotid.representations.gausscode.GaussCode`
        A pyknotid representation class for the knot, or anything that
        can automatically be converted into a GaussCode (i.e. by writing
        :code:`GaussCode(your_object)`), or a PlanarDiagram.
    '''
    from pyknotid.representations.gausscode import GaussCode
    from pyknotid.representations.planardiagram import PlanarDiagram
    if not isinstance(representation, (GaussCode, PlanarDiagram)):
        representation = GaussCode(representation)
    if isinstance(representation, GaussCode):
        representation = PlanarDiagram(representation)

    mathematica_code = representation.as_mathematica()
    _write_mathematica_script('jonesscript.m',
                              '<< KnotTheory`\nPrint[Jones[' +
                              mathematica_code + '][q]]')
    result = subprocess.check_output(['bash', 'jonesscript.m']).split('\n')[-2]
    q = sym.var('q')
    result = result.replace('[', '(')
    result = result.replace(']', ')')
    result = result.replace('Sqrt', 'sym.sqrt')
    r = re.compile('([0-9]+)')
    result = r.sub(r'sym.Integer(\1)', result)
    return eval(result.replace('^', '**'))

q = sym.var('q')
def jones_polynomial(representation, var=q):
    from pyknotid.representations.gausscode import GaussCode
    from pyknotid.representations.planardiagram import PlanarDiagram, Crossing
    if not isinstance(representation, (GaussCode, PlanarDiagram)):
        representation = GaussCode(representation)
    if isinstance(representation, GaussCode):
        representation = PlanarDiagram(representation)

    rep = representation

    if isinstance(q, sym.Symbol):
        A = sym.var('A')
    else:
        A = (var + 0j)**0.25
    
    crossings = crossings_connectedness_order(rep)
    if len(crossings) < 2:
        return 1
    
    web = []
    crossing = crossings[0]
    crossing = PlanarDiagram(crossing)
    diag_1, diag_2 = kauffman_skein(crossing)
    web.append([diag_1, A])
    web.append([diag_2, 1.0/A])
    replacement = -A**2 - A**-2

    for crossing in crossings[1:]:
        crossing = PlanarDiagram(crossing)
        diag_1, diag_2 = kauffman_skein(crossing)
        new_web = []

        for term in web:
            diag_1_product = PlanarDiagram(
                [x for x in diag_1] +
                [x for x in term[0]]).contract_points()
            diag_2_product = PlanarDiagram(
                [x for x in diag_2] +
                [x for x in term[0]]).contract_points()
            contracted_1 = 0
            contracted_2 = 0
            diag_1_new = []
            diag_2_new = []

            for point in diag_1_product:
                if len(point.components()) == 1:
                    contracted_1 += 1
                else:
                    diag_1_new.append(point)
            for point in diag_2_product:
                if len(point.components()) == 1:
                    contracted_2 += 1
                else:
                    diag_2_new.append(point)

            new_term_1 = ([sorted(diag_1_new)] +
                          [term[1] * (replacement ** contracted_1) * A])
            new_term_2 = ([sorted(diag_2_new)] +
                          [term[1] * (replacement ** contracted_2) * 1.0/A])

            new_web.append(new_term_1)
            new_web.append(new_term_2)

        unique_diags = list(set([tuple(x[0]) for x in new_web]))
        contracted_web = [[list(x)] for x in unique_diags]

        for term in new_web:
            for diag in contracted_web:
                if diag[0] == term[0]:
                    if len(diag) == 1:
                        diag.append(term[1])
                    else:
                        diag[1] += term[1]

        web = contracted_web

    # quick and dirty writhe calculation
    crossings = [x for x in rep if isinstance(x, Crossing)]
    total_writhe = 0
    for crossing in crossings:
        total_writhe += crossing.sign()
        
    jones_precursor = web[0][1]

    jones_poly = (-A**3)**total_writhe * jones_precursor / replacement

    if isinstance(var, sym.Symbol):
        return sym.expand(sym.simplify(jones_poly)).subs(A, var**0.25)
    return round(abs(jones_poly))


def crossings_connectedness_order(planar_diagram):
    from pyknotid.representations.planardiagram import Crossing
    crossings = [x for x in planar_diagram if isinstance(x, Crossing)]

    inner_components = crossings[0].components()
    crossings_remaining = crossings[1:]
    ordered_crossings = [crossings[0]]

    for i in range(len(crossings_remaining)):
        connectedness = []
        for crossing in crossings_remaining:
            new_components = (len(list(set(crossing.components() + inner_components))) -
                              len(inner_components))

            connectedness.append(len(crossing.components()) - new_components)
        most_connected_index = connectedness.index(max(connectedness))
        inner_components += crossings_remaining[most_connected_index].components()
        inner_components = list(set(inner_components))
        ordered_crossings.append(crossings_remaining[most_connected_index])
        del crossings_remaining[most_connected_index]

    return ordered_crossings

def contract_points(planar_diagram):
    '''
    For appropriately contracting :class: `Points` in a
    :class: `PlanarDiagram` According to the following rules:

    P_a,b P_b,c -> P_a,c
    P_a,b P_a,b -> P_a,a
    '''
    import copy

    pd_crossings = [x for x in planar_diagram if type(x) == Crossing]
    pd_points = [x for x in planar_diagram if type(x) == Point]

    pd_trimmed = PlanarDiagram()
    for crossing in pd_crossings:
        pd_trimmed.append(crossing)

    substituted_points = []
    for i in range(len(pd_points)):
        if pd_points[i] not in substituted_points:
            pd_trimmed.append(pd_points[i])
            totally_simplified = False
            while not totally_simplified:
                pd_trimmed_old = copy.copy(pd_trimmed)
                for j in range(i+1, len(pd_points)):
                    if (len(pd_trimmed[-1].components()) > 1 and
                            len(pd_points[j].components()) > 1 and
                            pd_points[j] not in substituted_points):
                        # This skips P_a,a situations
                        if pd_trimmed[-1].components() != pd_points[j].components():
                            # This conditions skips (P_a,b P_a,b) situations
                            substituted = False
                            p = 0
                            while substituted is not True and p != 2:
                                q = 0
                                while substituted is not True and q != 2:
                                    if pd_trimmed[-1].components()[p] == pd_points[j].components()[q]:
                                        pd_trimmed.append(Point(pd_trimmed[-1].components()[(p+1) % 2],
                                                                pd_points[j].components()[(q+1) % 2]))
                                        del pd_trimmed[-2]
                                        substituted = True
                                        substituted_points.append(pd_points[j])
                                    q += 1
                                p += 1
                        else:
                            # This deals with (P_a,b P_a,b) situations
                            pd_trimmed.append(Point(pd_trimmed[-1].components()[0],
                                                    pd_trimmed[-1].components()[0]))
                            del pd_trimmed[-2]
                            substituted_points.append(pd_points[j])
                if pd_trimmed_old == pd_trimmed:
                    totally_simplified = True
    return pd_trimmed


def _write_mathematica_script(filen, text):
    '''
    Write the given text (mathematica code) to the given filename. It will
    be wrapped in some MathKernel calling stuff first.
    '''
    with open(filen, 'w') as fileh:
        fileh.write('MathKernel -noprompt -run "commandLine={${1+\"$1\"}}; '
                    '$(sed \'1,/^exit/d\' $0) ; Exit[]"\nexit $?\n')
        fileh.write(text)
        fileh.close()


def _mathematica_matrix(cs, quadrant='lr', verbose=False):
    '''
    Turns the given crossings into a string of Mathematica code
    representing the Alexander matrix.

    This functions is for internal use only.
    '''
    if len(cs) == 0:
        return ''
    mathmat_entries = []

    line_num = 0
    num_crossings = int(len(cs)/2)
    crossing_num_counter = 0
    crossing_dict = {}
    crossing_exists = False
    written_indices = []
    for i, crossing in enumerate(cs):
        if verbose and (i+1) % 100 == 0:
            sys.stdout.write('\ri = {0} / {1}'.format(i, len(cs)))
        print('crossing is', crossing)
        identifier, upper, direc = crossing
        for entry in crossing_dict:
            if entry[0] == identifier:
                crossing_num = crossing_dict[entry]
                crossing_entry = entry
                crossing_exists = True
        if not crossing_exists:
            crossing_num = crossing_num_counter
            crossing_num_counter += 1
            crossing_dict[tuple(crossing)] = crossing_num
        else:
            crossing_dict.pop(crossing_entry)
        crossing_exists = False

        if upper > 0.99999:
            if direc > 0.99999:
                matrix_element = '1-1/t'
            else:
                matrix_element = '1-t'
            mathmat_entries.append((crossing_num,
                                    line_num % num_crossings,
                                    matrix_element))
            spec = (crossing_num, line_num % num_crossings)
            if spec in written_indices:
                print(spec)
            else:
                written_indices.append((crossing_num,
                                        line_num % num_crossings))
        else:
            if direc > 0.99999:
                new_matrix_element = '1/t'
            else:
                new_matrix_element = 't'
            mathmat_entries.append((crossing_num,
                                    line_num % num_crossings, '-1'))
            spec = (crossing_num, line_num % num_crossings)
            if spec in written_indices:
                print(spec)
            else:
                written_indices.append((crossing_num,
                                        line_num % num_crossings))
            line_num += 1
            mathmat_entries.append((crossing_num,
                                    line_num % num_crossings,
                                    new_matrix_element))
            spec = (crossing_num, line_num % num_crossings)
            if spec in written_indices:
                print(spec)
            else:
                written_indices.append((crossing_num,
                                        line_num % num_crossings))

    if quadrant == 'lr':
        mathmat_entries = filter(
            lambda j: j[0] != 0 and j[1] != 0, mathmat_entries)
        mathmat_entries = map(
            lambda j: (j[0]-1, j[1]-1, j[2]), mathmat_entries)
    if quadrant == 'ur':
        mathmat_entries = filter(
            lambda j: j[0] != num_crossings-1 and j[1] != 0, mathmat_entries)
        mathmat_entries = map(
            lambda j: (j[0], j[1]-1, j[2]), mathmat_entries)
    if quadrant == 'ul':
        mathmat_entries = filter(
            lambda j: j[0] != num_crossings-1 and j[1] != num_crossings-1,
            mathmat_entries)
    if quadrant == 'll':
        mathmat_entries = filter(
            lambda j: j[0] != 0 and j[1] != num_crossings-1,
            mathmat_entries)
        mathmat_entries = map(lambda j: (j[0]-1, j[1], j[2]), mathmat_entries)

    if verbose:
        print
    outstr = 'Sparse_array[{ '
    for entry in mathmat_entries:
        outstr += '{%d,%d}->%s, ' % (entry[0]+1, entry[1]+1, entry[2])
    outstr = outstr[:-2]
    outstr += ' }]'
    return outstr


def hyperbolic_volume(representation):
    '''
    The hyperbolic volume, calculated by the SnapPy library for
    studying the topology and geometry of 3-manifolds. This function
    depends on the Spherogram module, distributed with SnapPy or
    available separately.

    Parameters
    ----------
    representation : A PlanarDiagram, or anything convertible to a
                     :class:`~pyknotid.representations.gausscode.GaussCode`
        A pyknotid representation class for the knot, or anything that
        can automatically be converted into a GaussCode (i.e. by writing
        :code:`GaussCode(your_object)`), or a PlanarDiagram.

    '''
    from pyknotid.representations.gausscode import GaussCode
    from pyknotid.representations.planardiagram import PlanarDiagram
    if not isinstance(representation, (GaussCode, PlanarDiagram)):
        representation = GaussCode(representation)
    if isinstance(representation, GaussCode):
        representation = PlanarDiagram(representation)

    volume = representation.as_spherogram().exterior().volume()
    return volume

def _maxima_matrix(cs, quadrant='lr', verbose=False):
    '''
    Turns the given crossings into a string of maxima code
    representing the Alexander matrix.

    This functions is for internal use only.
    '''
    if len(cs) == 0:
        return ''
    mathmat_entries = {}

    line_num = 0
    num_crossings = int(len(cs)/2)
    crossing_num_counter = 0
    crossing_dict = {}
    crossing_exists = False
    written_indices = []
    for i, crossing in enumerate(cs):
        if verbose and (i+1) % 100 == 0:
            sys.stdout.write('\ri = {0} / {1}'.format(i, len(cs)))
        identifier, upper, direc = crossing
        for entry in crossing_dict:
            if entry[0] == identifier:
                crossing_num = crossing_dict[entry]
                crossing_entry = entry
                crossing_exists = True
        if not crossing_exists:
            crossing_num = crossing_num_counter
            crossing_num_counter += 1
            crossing_dict[tuple(crossing)] = crossing_num
        else:
            crossing_dict.pop(crossing_entry)
        crossing_exists = False

        if upper > 0.99999:
            if direc > 0.99999:
                matrix_element = '1-1/t'
            else:
                matrix_element = '1-t'
            mathmat_entries[
                (crossing_num, line_num % num_crossings)] =  matrix_element
            spec = (crossing_num, line_num % num_crossings)
            if spec in written_indices:
                print(spec)
            else:
                written_indices.append((crossing_num,
                                        line_num % num_crossings))
        else:
            if direc > 0.99999:
                new_matrix_element = '1/t'
            else:
                new_matrix_element = 't'
            mathmat_entries[(crossing_num, line_num % num_crossings)] =  '-1'
            spec = (crossing_num, line_num % num_crossings)
            if spec in written_indices:
                print(spec)
            else:
                written_indices.append((crossing_num,
                                        line_num % num_crossings))
            line_num += 1
            mathmat_entries[
                (crossing_num, line_num % num_crossings)] = new_matrix_element
            spec = (crossing_num, line_num % num_crossings)
            if spec in written_indices:
                print(spec)
            else:
                written_indices.append((crossing_num,
                                        line_num % num_crossings))

    if verbose:
        print
    outstrs = ['m: matrix(']
    num_crossings = int(len(cs) / 2)
    for row in range(num_crossings):
        outstrs.append('[')
        for column in range(num_crossings):
            if (row, column) in mathmat_entries:
                value = mathmat_entries[(row, column)]
            else:
                value = '0'
            outstrs.append(value)
            outstrs.append(', ')
        outstrs.pop()
        outstrs.append(']')
        outstrs.append(', ')
    outstrs.pop()
    outstrs.append(');\n')

    outstrs.append(
        {'lr': 'm2: submatrix(1, m, 1)',
         'ur': 'm2: submatrix({nc}, m, 1)'.format(nc=num_crossings),
         'ul': 'm2: submatrix({nc}, m, {nc})'.format(nc=num_crossings),
         'll': 'm2: submatrix(1, m, {nc})'.format(nc=num_crossings)
        }[quadrant])

    outstrs.append(';\n')
    return ''.join(outstrs)


def _cypari_matrix(cs, quadrant='lr', verbose=False):
    '''
    Turns the given crossings into a string of maxima code
    representing the Alexander matrix.

    This functions is for internal use only.
    '''
    if len(cs) == 0:
        return ''
    mathmat_entries = {}

    line_num = 0
    num_crossings = int(len(cs)/2)
    crossing_num_counter = 0
    crossing_dict = {}
    crossing_exists = False
    written_indices = []
    for i, crossing in enumerate(cs):
        if verbose and (i+1) % 100 == 0:
            sys.stdout.write('\ri = {0} / {1}'.format(i, len(cs)))
        identifier, upper, direc = crossing
        for entry in crossing_dict:
            if entry[0] == identifier:
                crossing_num = crossing_dict[entry]
                crossing_entry = entry
                crossing_exists = True
        if not crossing_exists:
            crossing_num = crossing_num_counter
            crossing_num_counter += 1
            crossing_dict[tuple(crossing)] = crossing_num
        else:
            crossing_dict.pop(crossing_entry)
        crossing_exists = False

        if upper > 0.99999:
            if direc > 0.99999:
                matrix_element = '1-1/t'
            else:
                matrix_element = '1-t'
            mathmat_entries[
                (crossing_num, line_num % num_crossings)] =  matrix_element
            spec = (crossing_num, line_num % num_crossings)
            if spec in written_indices:
                print(spec)
            else:
                written_indices.append((crossing_num,
                                        line_num % num_crossings))
        else:
            if direc > 0.99999:
                new_matrix_element = '1/t'
            else:
                new_matrix_element = 't'
            mathmat_entries[(crossing_num, line_num % num_crossings)] =  '-1'
            spec = (crossing_num, line_num % num_crossings)
            if spec in written_indices:
                print(spec)
            else:
                written_indices.append((crossing_num,
                                        line_num % num_crossings))
            line_num += 1
            mathmat_entries[
                (crossing_num, line_num % num_crossings)] = new_matrix_element
            spec = (crossing_num, line_num % num_crossings)
            if spec in written_indices:
                print(spec)
            else:
                written_indices.append((crossing_num,
                                        line_num % num_crossings))

    if verbose:
        print

    print('Warning: quadrant ignored!')
    
    outstrs = ['[']
    num_crossings = int(len(cs) / 2)
    for row in range(num_crossings-1):
        for column in range(num_crossings-1):
            if (row, column) in mathmat_entries:
                value = mathmat_entries[(row, column)]
            else:
                value = '0'
            outstrs.append(value)
            outstrs.append(', ')
        outstrs.pop()
        outstrs.append(';')
    outstrs.pop()
    outstrs.append(']')

    from cypari.gen import pari
    return pari(''.join(outstrs))

def _crossing_arrows_and_signs(gc, crossing_numbers):
    '''Internal function to get a list of crossing arrow indices
    (as per a Gauss diagram) and signs, both in dictionaries.'''

    ## Arrow diagram for v_2 is
    ## appear -> leave -> leave -> appear (with crossed arrows)
    over_crossing_indices = {}
    under_crossing_indices = {}
    signs = {}
    for i, row in enumerate(gc):
        if row[1] == 1:
            over_crossing_indices[row[0]] = i
        else:
            under_crossing_indices[row[0]] = i
        signs[row[0]] = row[2]

    arrows = {index: (over_crossing_indices[index],
                      under_crossing_indices[index])
              for index in crossing_numbers}

    return arrows, signs

def _crossing_arrows_and_signs_numpy(gc, crossing_numbers):
    ## Arrow diagram for v_2 is
    ## appear -> leave -> leave -> appear (with crossed arrows)
    over_crossing_indices = {}
    under_crossing_indices = {}
    signs = {}
    for i, row in enumerate(gc):
        if row[1] == 1:
            over_crossing_indices[row[0]] = i
        else:
            under_crossing_indices[row[0]] = i
        signs[row[0]] = row[2]

    arrows = n.zeros((len(crossing_numbers), 3), dtype=n.long)

    for index, number in enumerate(crossing_numbers):
        row = arrows[index]
        row[0] = over_crossing_indices[number]
        row[1] = under_crossing_indices[number]
        row[2] = signs[number]

    return arrows

def second_order_writhe(representation):
    '''Returns the second order writhe (i1,i3,i2,i4) of the
    representation, as defined in Lin and Wang.
    '''

    from pyknotid.representations.gausscode import GaussCode
    if not isinstance(representation, GaussCode):
        representation = GaussCode(representation)

    gc = representation._gauss_code
    if len(gc) == 0:
        return 0
    elif len(gc) > 1:
        raise Exception('tried to calculate second order writhe'
                        'for something with more than 1 component')

    gc = gc[0].copy()

    arrows, signs = _crossing_arrows_and_signs(
        gc, representation.crossing_numbers)
    
    crossing_numbers = list(representation.crossing_numbers)
    result = 0
    for index, i1 in enumerate(crossing_numbers):
        arrow1 = arrows[i1]
        a1s, a1e = arrow1
        for i2 in crossing_numbers[index+1:]:
            arrow2 = arrows[i2]
            a2s, a2e = arrow2

            if ((a2s > a1e and a1s > a2s and a2e > a1s) or
                (a2e > a1e and a1s > a2e and a2s > a1s) or
                (a2e > a1s and a1e > a2e and a2s > a1e) or
                (a2s > a1s and a1e > a2s and a2e > a1e)):
            # if ((a2e > a1e and a1s > a2e and a2s > a1s) or
            #     (a2s > a1s and a1e > a2s and a2e > a1e)):
                result += signs[i1] * signs[i2]

    return result

def arnold_2St_2Jplus(representation):
    '''Returns J+ + 2 * St where J+ and St are Arnold's invariants of
    plane curves.

    The calculation is performed by transforming the representation
    into a representation of an unknot by flipping crossings, then
    calculating the second order writhe.

    See 'Invariants of curves and fronts via Gauss diagrams', M
    Polyak, Topology 37, 1998.

    '''

    from pyknotid.representations.gausscode import GaussCode
    if not isinstance(representation, GaussCode):
        representation = GaussCode(representation)

    gc = representation._gauss_code
    if len(gc) == 0:
        return 0
    elif len(gc) > 1:
        raise Exception('tried to calculate arnold_2St_2Jplus'
                        'for something with more than 1 component')

    gc = gc[0].copy()
    seen_crossings = set()
    changed_crossings = set()
    for row in gc:
        if row[0] not in seen_crossings:
            seen_crossings.add(row[0])
            if row[1] < 0:
                changed_crossings.add(row[0])
                row[1] *= -1
                row[2] *= -1
        elif row[0] in changed_crossings:
            row[1] *= -1
            row[2] *= -1
    arrows, signs = _crossing_arrows_and_signs(
        gc, representation.crossing_numbers)
    
    crossing_numbers = list(representation.crossing_numbers)
    result = 0
    for index, i1 in enumerate(crossing_numbers):
        arrow1 = arrows[i1]
        a1s, a1e = arrow1
        for i2 in crossing_numbers[index+1:]:
            arrow2 = arrows[i2]
            a2s, a2e = arrow2

            if ((a2s > a1e and a1s > a2s and a2e > a1s) or
                (a2e > a1e and a1s > a2e and a2s > a1s) or
                (a2e > a1s and a1e > a2e and a2s > a1e) or
                (a2s > a1s and a1e > a2s and a2e > a1e)):
            # if ((a2e > a1e and a1s > a2e and a2s > a1s) or
            #     (a2s > a1s and a1e > a2s and a2e > a1e)):
                result += signs[i1] * signs[i2]

    return result

def arnold_2St_2Jminus(representation):
    '''Returns J- + 2 * St where J+ and St are Arnold's invariants of
    plane curves.

    See 'Invariants of curves and fronts via Gauss diagrams', M
    Polyak, Topology 37, 1998.

    '''
    return arnold_2St_2Jplus(representation) - len(representation)

def vassiliev_degree_2(representation):
    '''Calculates the Vassiliev invariant of degree 2 of the given
    knot. The representation must have just one knot component,
    this doesn't work for links.

    Parameters
    ==========
    representation : Anything convertible to a
                     :class:`~pyknotid.representations.gausscode.GaussCode`
        A pyknotid representation class for the knot, or anything that
        can automatically be converted into a GaussCode (i.e. by writing
        :code:`GaussCode(your_object)`).
    '''
    ## See Polyak and Viro
    from pyknotid.representations.gausscode import GaussCode
    if not isinstance(representation, GaussCode):
        representation = GaussCode(representation)

    gc = representation._gauss_code
    if len(gc) == 0:
        return 0
    elif len(gc) > 1:
        raise Exception('tried to calculate v2'
                        'for something with more than 1 component')

    gc = gc[0]
    arrows, signs = _crossing_arrows_and_signs(
        gc, representation.crossing_numbers)
    
    crossing_numbers = list(representation.crossing_numbers)
    representations_sum = 0
    for index, i1 in enumerate(crossing_numbers):
        arrow1 = arrows[i1]
        a1s, a1e = arrow1
        for i2 in crossing_numbers[index+1:]:
            arrow2 = arrows[i2]
            a2s, a2e = arrow2

            if a2s > a1s and a2e < a1s and a1e > a2s:
                representations_sum += signs[i1] * signs[i2]
            elif a1s > a2s and a1e < a2s and a2e > a1s:
                representations_sum += signs[i1] * signs[i2]

    return representations_sum

def vassiliev_degree_3(representation, try_cython=True):
    '''Calculates the Vassiliev invariant of degree 3 of the given
    knot. The representation must have just one knot component,
    this doesn't work for links.

    Parameters
    ----------
    representation : Anything convertible to a
                     :class:`~pyknotid.representations.gausscode.GaussCode`
        A pyknotid representation class for the knot, or anything that
        can automatically be converted into a GaussCode (i.e. by writing
        :code:`GaussCode(your_object)`).
    try_cython : bool
        Whether to try and use an optimised cython version of the
        routine (takes about 1/3 of the time for complex representations).
        Defaults to True, but the python fallback will be *slower*
        than setting it to False if the cython function is not
        available.
    '''

    if try_cython:
        return _vassiliev_degree_3_numpy(representation)
    return _vassiliev_degree_3_python(representation)

        
def _vassiliev_degree_3_python(representation):
    ## See Polyak and Viro
    from pyknotid.representations.gausscode import GaussCode
    if not isinstance(representation, GaussCode):
        representation = GaussCode(representation)

    gc = representation._gauss_code
    if len(gc) == 0:
        return 0
    elif len(gc) > 1:
        raise Exception('tried to calculate alexander polynomial'
                        'for something with more than 1 component')

    gc = gc[0]
    arrows, signs = _crossing_arrows_and_signs(
        gc, representation.crossing_numbers)
    
    crossing_numbers = list(representation.crossing_numbers)
    used_sets = set()
    representations_sum_1 = 0
    representations_sum_2 = 0
    for index, i1 in enumerate(crossing_numbers):
        if index % 10 == 0:
            vprint('\rCurrently comparing index {}'.format(index), False)
        arrow1 = arrows[i1]
        a1s, a1e = arrow1
        a1e = (a1e - a1s) % len(gc)
        for i2 in crossing_numbers:
            arrow2 = arrows[i2]
            a2s, a2e = arrow2
            a2s = (a2s - a1s) % len(gc)
            a2e = (a2e - a1s) % len(gc)
            for i3 in crossing_numbers:
                arrow3 = arrows[i3]
                a3s, a3e = arrow3
                a3s = (a3s - a1s) % len(gc)
                a3e = (a3e - a1s) % len(gc)

                ordered_indices = tuple(sorted((i1, i2, i3)))
                if ordered_indices in used_sets:
                    continue

                if (a2s < a1e and a3e < a1e and a3e > a2s and
                    a3s > a1e and a2e > a3s):
                    representations_sum_1 += (signs[i1] * signs[i2] *
                                              signs[i3])
                    used_sets.add(ordered_indices)
                if (a2e < a1e and a3s < a1e and a3s > a2e and
                    a2s > a1e and a3e > a2s):
                    representations_sum_2 += (signs[i1] * signs[i2] *
                                              signs[i3])
                    used_sets.add(ordered_indices)

    print()
    
    return int(round(representations_sum_1 / 2.)) + representations_sum_2

    
def _vassiliev_degree_3_numpy(representation):
    ## See Polyak and Viro
    from pyknotid.representations.gausscode import GaussCode
    if not isinstance(representation, GaussCode):
        representation = GaussCode(representation)

    gc = representation._gauss_code
    if len(gc) == 0:
        return 0
    elif len(gc) > 1:
        raise Exception('tried to calculate alexander polynomial'
                        'for something with more than 1 component')

    gc = gc[0]
    arrows = _crossing_arrows_and_signs_numpy(
        gc, representation.crossing_numbers)

    try:
        from pyknotid import cinvariants
    except ImportError:
        print('Failed to import cinvariants. Using *slow* python numpy method.')
    else:
        return int(round(cinvariants.vassiliev_degree_3(arrows)))

    num_crossings = len(arrows) * 2
    
    used_sets = set()
    representations_sum_1 = 0
    representations_sum_2 = 0
    arrow_range = range(len(arrows))
    for i1 in arrow_range:
        if i1 % 10 == 0:
            vprint('\rCurrently comparing index {}   '.format(i1), False)
        arrow1 = arrows[i1]
        a1s, a1e, sign1 = arrow1
        a1e = (a1e - a1s) % num_crossings
        for i2 in arrow_range:
            arrow2 = arrows[i2]
            a2s, a2e, sign2 = arrow2
            a2s = (a2s - a1s) % num_crossings
            a2e = (a2e - a1s) % num_crossings
            for i3 in arrow_range:
                arrow3 = arrows[i3]
                a3s, a3e, sign3 = arrow3
                a3s = (a3s - a1s) % num_crossings
                a3e = (a3e - a1s) % num_crossings

                ordered_indices = tuple(sorted((i1, i2, i3)))
                if ordered_indices in used_sets:
                    continue

                if (a2s < a1e and a3e < a1e and a3e > a2s and
                    a3s > a1e and a2e > a3s):
                    representations_sum_1 += sign1 * sign2 * sign3
                    used_sets.add(ordered_indices)
                if (a2e < a1e and a3s < a1e and a3s > a2e and
                    a2s > a1e and a3e > a2s):
                    representations_sum_2 += sign1 * sign2 * sign3
                    used_sets.add(ordered_indices)

    print
                    
    
    return int(round(representations_sum_1 / 2.)) + representations_sum_2


def self_linking(representation):
    '''Returns the self linking number J(K) of the Gauss code, an
    invariant of virtual knots. See Kauffman 2004 for more
    information.

    Currently only works for knots.
    '''

    from pyknotid.representations.gausscode import GaussCode
    if not isinstance(representation, GaussCode):
        representation = GaussCode(representation)

    if len(representation) == 0:
        return 0
    if len(representation._gauss_code[0]) == 0:
        return 0
    gauss_code = representation._gauss_code
    l = len(gauss_code[0][:,0])
    total_crossings = l/2
    crossing_counter = 1
    slink_counter = 0        

    for crossing_number in representation.crossing_numbers:
        occurences = n.where(gauss_code[0][:, 0] == crossing_number)[0]
        first_occurence = occurences[0]
        second_occurence = occurences[1]
        crossing_difference = second_occurence - first_occurence        

        if(crossing_difference % 2 == 0):
            slink_counter += 2 * gauss_code[0][occurences[0],2]

    return slink_counter   

def virtual_vassiliev_degree_3(representation):
    '''Calculates the virtual Vassiliev invariant of degree 3 (for
    non-long-knots) of the given representation, as described in
    'Finite type invariants of classical and virtual knots' by
    Goussarov, Polyak and Viro.

    Parameters
    ----------
    representation : :class:`~pyknotid.representations.representation.Representation`
        A representation class, or anything convertible to one
        (in principle).
    '''
    ## See Polyak and Viro and Goussarov
    from pyknotid.representations.gausscode import GaussCode
    if not isinstance(representation, GaussCode):
        representation = GaussCode(representation)

    gc = representation._gauss_code
    if len(gc) == 0:
        return 0
    elif len(gc) > 1:
        raise Exception('tried to calculate alexander polynomial'
                        'for something with more than 1 component')

    gc = gc[0]
    arrows, signs = _crossing_arrows_and_signs(
        gc, representation.crossing_numbers)

    diagrams_found = [[] for _ in range(8)]
    
    crossing_numbers = list(representation.crossing_numbers)
    used_sets = set()
    representations_sum = 0
    for index, i1 in enumerate(crossing_numbers):
        if index % 10 == 0 and len(crossing_numbers) > 10:
            vprint('\rCurrently comparing index {}'.format(index), False)
        arrow1 = arrows[i1]
        a1s, a1e = arrow1
        a1e = (a1e - a1s) % len(gc)
        for i2 in crossing_numbers:
            arrow2 = arrows[i2]
            a2s, a2e = arrow2
            a2s = (a2s - a1s) % len(gc)
            a2e = (a2e - a1s) % len(gc)

            # For the virtual Vassilev of order f, there are still
            # some diagrams with just 2 arrows:
            first_two_ordered_indices = tuple(sorted((i1, i2)))
            if (a2s < a1e and a2e > a1e and
                signs[i1] == signs[i2] and
                first_two_ordered_indices not in used_sets):
                if signs[i1] > 0:
                    representations_sum -= 1
                    # print('7')
                else:
                    representations_sum += 1
                    # print('8')
                used_sets.add(first_two_ordered_indices)
           
            for i3 in crossing_numbers:
                arrow3 = arrows[i3]
                a3s, a3e = arrow3
                a3s = (a3s - a1s) % len(gc)
                a3e = (a3e - a1s) % len(gc)

                ordered_indices = tuple(sorted((i1, i2, i3)))
                if ordered_indices in used_sets:
                    continue

                # These checks are for the diagrams of Polyak, Viro
                # and Goussarov in 'Finite type invariants of
                # classical and virtual knots' (page 13)

                if (a2e < a1e and a3s < a1e and a3s > a2e and
                    a2s > a1e and a3e > a2s):
                    representations_sum += 3*(signs[i1] * signs[i2] *
                                              signs[i3])
                    used_sets.add(ordered_indices)
                    # print('1')

                if (a2s < a1e and a3s < a1e and a2s < a3s and
                    a2e < a3e and a2e > a1e):
                    representations_sum -= (signs[i1] * signs[i2] *
                                              signs[i3])
                    used_sets.add(ordered_indices)
                    # print('2')

                if (a2s < a1e and a3e < a1e and a3e > a2s and
                    a3s > a1e and a2e > a3s):
                    representations_sum += (signs[i1] * signs[i2] *
                                              signs[i3])
                    used_sets.add(ordered_indices)
                    # print('3')

                if (a2e < a1e and a3s < a1e and a3s > a2e and
                    a3e > a1e and a2s > a3e):
                    representations_sum += (signs[i1] * signs[i2] *
                                              signs[i3])
                    used_sets.add(ordered_indices)
                    # print('4')

                if (a2e < a1e and a3e < a1e and a3s < a2s and
                    a3s > a1e):
                    representations_sum -= (signs[i1] * signs[i2] *
                                              signs[i3])
                    used_sets.add(ordered_indices)
                    # print('5')

                if (a2s < a1e and a3s < a1e and a3e < a2e and
                    a3e > a1e):
                    representations_sum -= (signs[i1] * signs[i2] *
                                              signs[i3])
                    used_sets.add(ordered_indices)
                    # print('6')

    print()

    return representations_sum
    return int(round(representations_sum))




from __future__ import division, print_function
import numpy as np
from math import factorial
from functools import wraps
from itertools import combinations, permutations
from collections import defaultdict

from pyknotid.representations.gausscode import GaussCode
from pyknotid.utils import vprint


def _to_GaussCode(rep):
    if not isinstance(rep, GaussCode):
        rep = GaussCode(rep)
    return rep


def require_GaussCode(func):
    '''Convert the first argument of the function to a GaussCode.'''
    @wraps(func)
    def wrapper(rep, *args, **kwargs):
        rep = _to_GaussCode(rep)
        return func(rep, *args, **kwargs)
    return wrapper


def validate_diagram(d):
    '''Do some basic checks on whether an Arrow diagram (as a string of
    numbered crossings and signs) is valid.'''
    entries = d.split(',')
    counts = defaultdict(lambda: 0)
    signs = defaultdict(lambda: 0)
    for entry in entries:
        number = entry[:-1]
        if entry[-1] not in ('+', '-'):
            raise ValueError('Arrow diagrams must have crossing + or - ',
                             'but this has {}'.format(entry))
        counts[number] += 1
        signs[number] += (1 if entry[-1] == '+' else -1)

    if any([item != 2 for item in counts.values()]):
        raise ValueError('Diagram {} appears invalid: some indices '
                         'appear only once'.format(d))

    if any([item != 0 for item in signs.values()]):
        raise ValueError('Diagram {} appears invalid: sum of signs for '
                         'some crossings do not equal 0'.format(d))
        

@require_GaussCode
def writhing_numbers(gc, diagrams, based=False):
    '''Returns the signed sum of representations of the given Arrow
    diagrams in the given representation.

    Parameters
    ----------

    gc : A :class:`~pyknotid.representations.gausscode.GaussCode` or
         equivalent representation.
        The knot for which to find the writhes. 
    diagrams : str or list or tuple
        A list of strings, or single string, representing Arrow
        diagrams, e.g. '1-2+1+2-' for Vassiliev 2.
    based : bool
        Whether the diagrams have basepoints (if True, assumed to be
        just before the first entry).
    '''

    if not isinstance(diagrams, (list, tuple)):
        diagrams = [diagrams]

    multipliers = defaultdict(lambda : {})
    use_multipliers = False
    if any(['d' in diagram for diagram in diagrams]):
        use_multipliers = True
    if use_multipliers:
        for di, diagram  in enumerate(diagrams):
            terms = diagram.split(',')
            for term in terms:
                if term[-1] == 'd':
                    multiplier = 2
                else:
                    multiplier = 1
                term = int(term[:-2]) if 'd' in term else int(term[:-1])
                multipliers[diagram.replace('d', '')][term] = multiplier
            diagrams[di] = diagram.replace('d', '')

    for d in diagrams:
        validate_diagram(d)

    level = 0

    code = gc._gauss_code
    code = code[0]
    gc_len = len(gc)
    code_len = len(code)
    from pyknotid.invariants import _crossing_arrows_and_signs
    arrows, signs = _crossing_arrows_and_signs(code, gc.crossing_numbers)

    crossing_numbers = list(gc.crossing_numbers)

    # degrees = [len(diagram.split(',')) for diagram in diagrams]

    degrees = defaultdict(lambda: [])
    for diagram in diagrams:
        degrees[len(diagram.split(',')) // 2].append(diagram)

    relations = {diagram: [] for diagram in diagrams}
    for diagram in diagrams:
        degree = len(diagram.split(',')) // 2
        num_relations = factorial(degree - 1) * 4

        terms = diagram.split(',')
        numbers = [term[:-1] for term in terms]

        number_strs = list(sorted(set(numbers), key=lambda j: int(j)))
        for i, number in enumerate(number_strs):
            for oi, other_number in enumerate(number_strs[i+1:]):
                oi += i + 1
                if i != 0:
                    if terms.index(number + '-') < terms.index(other_number + '-'):
                        relations[diagram].append(lambda l, i=i, oi=oi: l[i][0] < l[oi][0])
                    else:
                        relations[diagram].append(lambda l, i=i, oi=oi: l[i][0] > l[oi][0])

                if terms.index(number + '-') < terms.index(other_number + '+'):
                    relations[diagram].append(lambda l, i=i, oi=oi: l[i][0] < l[oi][1])
                else:
                    relations[diagram].append(lambda l, i=i, oi=oi: l[i][0] > l[oi][1])

                if terms.index(number + '+') < terms.index(other_number + '-'):
                    relations[diagram].append(lambda l, i=i, oi=oi: l[i][1] < l[oi][0])
                else:
                    relations[diagram].append(lambda l, i=i, oi=oi: l[i][1] > l[oi][0])

                # This one seems to not be necessary for diagrams where all arrows cross?
                if terms.index(number + '+') < terms.index(other_number + '+'):
                    relations[diagram].append(lambda l, i=i, oi=oi: l[i][1] < l[oi][1])
                else:
                    relations[diagram].append(lambda l, i=i, oi=oi: l[i][1] > l[oi][1])

            if i == 0:
                continue
            if terms.index(number + '+') < terms.index(number + '-'):
                relations[diagram].append(lambda l, i=i, oi=oi: l[i][1] < l[i][0])
            else:
                relations[diagram].append(lambda l, i=i, oi=oi: l[i][1] > l[i][0])


    max_degree = max(degrees.keys())

    representations_sums = {d: 0 for d in diagrams}
    used_sets = {d: set() for d in diagrams}

    combs = combinations(crossing_numbers, max_degree)
    try:
        num_combs = (factorial(len(crossing_numbers)) //
                     factorial(max_degree) //
                     factorial(len(crossing_numbers) - max_degree))
    except ValueError:
        num_combs = 0
    strs = [None for _ in range(max_degree)] * 2
    order = [None for _ in range(max_degree)] * 2
    for ci, comb in enumerate(combs):
        if ci % 10000 == 0:
            vprint('\rCombination {} of {}    '.format(ci + 1, num_combs),
                   newline=False, condition=(ci % 10000) == 0)

        if based:
            perms = [comb]
        else:
            perms = permutations(comb)

        ordered_indices = tuple(sorted(comb))
        for diagram in diagrams:
            if ordered_indices not in used_sets[diagram]:
                break
        else:
            continue

        for perm in perms:
            cur_arrows = [list(arrows[i]) for i in perm]

            a1s = cur_arrows[0][0]
            if based:
                a1s = 0

            for i, arrow in enumerate(cur_arrows):
                arrow[0] = (arrow[0] - a1s) % code_len
                arrow[1] = (arrow[1] - a1s) % code_len


            for diagram in diagrams:
                if ordered_indices in used_sets[diagram]:
                    continue
                for relation in relations[diagram]:
                    if not relation(cur_arrows):
                        break
                else:
                    if use_multipliers:
                        representations_sums[diagram] += (
                            reduce(lambda x, y: x*y,
                                   [signs[arrow_i]**multipliers[diagram][num+1] for num, arrow_i in enumerate(perm)]))
                    else:
                        representations_sums[diagram] += (
                            reduce(lambda x, y: x*y,
                                   [signs[arrow_i] for arrow_i in perm]))
                    used_sets[diagram].add(ordered_indices)
                        
    vprint()

    return representations_sums


def vassiliev_2(gc):
    results = writhing_numbers(gc, '1-,2+,1+,2-', based=True)
    print('results', results)
    return results['1-,2+,1+,2-']

def vassiliev_3(gc):
    results =  writhing_numbers(gc, ['1-,2+,3-,1+,2-,3+',
                                     '1-,2-,3+,1+,3-,2+'], based=False,
                                forbid_all_symmetry=True)
    print('results', results)
    return results['1-,2-,3+,1+,3-,2+'] // 2 + results['1-,2+,3-,1+,2-,3+']

def vassiliev_4(gc):
    d4 = ['1-,2-,3+,1+,4-,3-,2+,4+',
          '1-,2+,3-,1+,4-,3+,2-,4+',
          '1-,2+,3-,4+,2-,1+,4-,3+',
          '1-,2-,3+,1+,4-,2+,3-,4+',
          '1-,2+,3-,4+,1+,2-,4-,3+',
          '1-,2+,3-,4-,1+,4+,3+,2-',
          '1-,2+,3-,4-,1+,3+,4+,2-',
          '1-,2+,3-,4+,1+,4-,3+,2-',
          '1-,2+,3+,4-,1+,3-,4+,2-',
          '1-,2+,3-,4-,1+,4+,2-,3+',
          '1-,2-,3+,4-,2+,1+,3-,4+',
          '1-,2-,3+,4-,1+,2+,3-,4+',
          '1-,2-,1+,3-,2+,4-,3+,4+',
          '1-,2-,1+,3+,2+,4-,3-,4+']

    d3 = ['1-,2+,3-d,1+,3+d,2-',
          '1-d,2+,3-,1+d,2-,3+']

    results_4 = writhing_numbers(gc, d4)

    results_3 = writhing_numbers(gc, d3)

    total = (results_4[d4[0]] + 6*results_4[d4[1]] + 2*results_4[d4[2]] +
             3*results_4[d4[3]] + results_4[d4[4]] + 2*results_4[d4[5]] +
             2*results_4[d4[6]] - results_4[d4[7]] + results_4[d4[8]] +
             results_4[d4[9]] + 2*results_4[d4[10]] + 2*results_4[d4[11]] +
             results_4[d4[12]] + results_4[d4[13]] +
             results_3[d3[0]] + results_3[d3[1]])

    print('results 4')
    for diagram, value in results_4.items():
        print(diagram, ':', value)
    print('results 3')
    for diagram, value in results_3.items():
        print(diagram, ':', value)
    return total

             
def slip_vassiliev_2(gc):
    codes = ['1-,2+,3+,1+,2-,3-',
             '1-,2+,3+,1+,3-,2-',
             '1-,2-,3+,1+,2+,3-',
             '1-,2-,3+,2+,1+,3-']
    # codes = ['1-,2+,3+,1+,2-,3-']
    results = writhing_numbers(gc, codes, based=True)

    gc = gc.flipped()
    results2 = writhing_numbers(gc, codes, based=True)

    for code in codes:
        print('{}: {}'.format(code, results[code]))
    return np.sum(results.values()) + np.sum(results2.values())
    return np.sum(results.values()), np.sum(results2.values())


def vassiliev_2_long_form(gc):

    results_1 = writhing_numbers(gc, ['1-,2+,1+,2-'], based=False)

    results_2 = writhing_numbers(gc, ['1-,2-d,3-,1+,2+d,3+',
                                      '1-,2-,3-d,1+,3+d,2+',
                                      '1-d,2-,3-,2+,1+d,3+',
                                      '1-,2-,1+,3-d,3+d,2+',
                                      '1-,2-,1+,3+d,3-d,2+'],
                                 based=False)
    
    num_crossings = float(len(gc.crossing_numbers))

    total = np.sum(results_2.values()) + results_1.values()[0]
    if total != 0 and num_crossings == 0:
        raise ValueError('Found no crossings but non-zero total vassiliev')
    if num_crossings == 0:
        return 0

    return total / num_crossings


    


@require_GaussCode
def writhing_numbers_numpy(gc, diagrams, based=False):
    '''Returns the signed sum of representations of the given Arrow
    diagrams in the given representation.

    Parameters
    ----------

    gc : A :class:`~pyknotid.representations.gausscode.GaussCode` or
         equivalent representation.
        The knot for which to find the writhes. 
    diagrams : str or list or tuple
        A list of strings, or single string, representing Arrow
        diagrams, e.g. '1-2+1+2-' for Vassiliev 2.
    based : bool
        Whether the diagrams have basepoints (if True, assumed to be
        just before the first entry).
    '''

    if not isinstance(diagrams, (list, tuple)):
        diagrams = [diagrams]

    for d in diagrams:
        validate_diagram(d)

    level = 0

    code = gc._gauss_code
    code = code[0]
    gc_len = len(gc)
    code_len = len(code)
    from pyknotid.invariants import _crossing_arrows_and_signs
    arrows, signs = _crossing_arrows_and_signs(code, gc.crossing_numbers)

    crossing_numbers = list(gc.crossing_numbers)

    # degrees = [len(diagram.split(',')) for diagram in diagrams]

    degrees = defaultdict(lambda: [])
    for diagram in diagrams:
        degrees[len(diagram.split(',')) // 2].append(diagram)

    relations = {diagram: [] for diagram in diagrams}
    for diagram in diagrams:
        degree = len(diagram.split(',')) // 2
        num_relations = factorial(degree - 1) * 4

        terms = diagram.split(',')
        numbers = [term[:-1] for term in terms]

        number_strs = list(sorted(set(numbers), key=lambda j: int(j)))
        for i, number in enumerate(number_strs):
            for oi, other_number in enumerate(number_strs[i+1:]):
                oi += i + 1
                if i != 0:
                    if terms.index(number + '-') < terms.index(other_number + '-'):
                        relations[diagram].append(lambda l, i=i, oi=oi: l[i][0] < l[oi][0])
                    else:
                        relations[diagram].append(lambda l, i=i, oi=oi: l[i][0] > l[oi][0])

                    if terms.index(number + '-') < terms.index(other_number + '+'):
                        relations[diagram].append(lambda l, i=i, oi=oi: l[i][0] < l[oi][1])
                    else:
                        relations[diagram].append(lambda l, i=i, oi=oi: l[i][0] > l[oi][1])

                if terms.index(number + '+') < terms.index(other_number + '-'):
                    relations[diagram].append(lambda l, i=i, oi=oi: l[i][1] < l[oi][0])
                else:
                    relations[diagram].append(lambda l, i=i, oi=oi: l[i][1] > l[oi][0])

                if i == 0:
                    if terms.index(number + '+') < terms.index(other_number + '+'):
                        relations[diagram].append(lambda l, i=i, oi=oi: l[i][1] < l[oi][1])
                    else:
                        relations[diagram].append(lambda l, i=i, oi=oi: l[i][1] > l[oi][1])


    max_degree = max(degrees.keys())

    used_sets = set()

    # representations_sums = [0 for _ in diagrams]
    representations_sums = {d: 0 for d in diagrams}
    used_sets = {d: set() for d in diagrams}

    print('arrows are', arrows)

    combs = combinations(crossing_numbers, max_degree)
    try:
        num_combs = (factorial(len(crossing_numbers)) //
                     factorial(max_degree) //
                     factorial(len(crossing_numbers) - max_degree))
    except ValueError:
        num_combs = 0
    strs = [None for _ in range(max_degree)] * 2
    order = [None for _ in range(max_degree)] * 2
    for ci, comb in enumerate(combs):
        if ci % 10000 == 0:
            vprint('\rCombination {} of {}    '.format(ci + 1, num_combs),
                   newline=False)

        if based:
            perms = [comb]
        else:
            perms = permutations(comb)

        ordered_indices = tuple(sorted(comb))
        for diagram in diagrams:
            if ordered_indices not in used_sets[diagram]:
                break
        else:
            continue

        for perm in perms:
            cur_arrows = [list(arrows[i]) for i in perm]

            a1s = cur_arrows[0][0]
            if based:
                a1s = 0

            for i, arrow in enumerate(cur_arrows):
                arrow[0] = (arrow[0] - a1s) % code_len
                arrow[1] = (arrow[1] - a1s) % code_len


            for diagram in diagrams:
                if ordered_indices in used_sets[diagram]:
                    continue
                for relation in relations[diagram]:
                    if not relation(cur_arrows):
                        break
                else:
                    representations_sums[diagram] += (
                        reduce(lambda x, y: x*y,
                               [signs[arrow_i] for arrow_i in perm]))
                    used_sets[diagram].add(ordered_indices)
                        
    vprint()

    return representations_sums



'''
OpenKnot
========

Class for working with open (linear) curves, that do not form closed
loops. :class:`OpenKnot` provides methods for visualising these curves
and analysing their topology via different kinds of closures.

API documentation
~~~~~~~~~~~~~~~~~

'''

from __future__ import print_function
import numpy as n
import sympy as sym

from pyknotid.spacecurves.spacecurve import SpaceCurve
from pyknotid.spacecurves.knot import Knot
from pyknotid.spacecurves.rotation import get_rotation_angles, rotate_to_top
from collections import Counter

from pyknotid.invariants import alexander
from pyknotid.representations.gausscode import GaussCode
from pyknotid.representations.representation import Representation
from pyknotid.visualise import plot_shell, plot_sphere_shell_vispy


class OpenKnot(SpaceCurve):
    '''
    Class for holding the vertices of a single line that is assumed to
    be an open curve. This class inherits from
    :class:`~pyknotid.spacecurves.spacecurve.SpaceCurve`, replacing any
    default arguments that assume closed curves, and providing methods
    for statistical analysis of knot invariants on projection and closure.

    All knot invariant methods return the results of a sampling over
    many projections of the knot, unless indicated otherwise.
    '''

    def __init__(self, *args, **kwargs):
        super(OpenKnot, self).__init__(*args, **kwargs)
        self._cached_v2 = {}
        self._cached_v3 = {}

    @SpaceCurve.points.setter
    def points(self, points):
        super(OpenKnot, self.__class__).points.fset(self, points)
        self._cached_alexanders = None

    def tangents(self):
        return super(Knot, self).tangents(closed=True)

    def closing_distance(self):
        '''
        Returns the distance between the first and last points.
        '''
        return n.linalg.norm(self.points[-1] - self.points[0])

    def raw_crossings(self, mode='use_max_jump',
                      virtual_closure=False,
                      recalculate=False, try_cython=False):
        '''
        Calls :meth:`pyknotid.spacecurves.spacecurve.SpaceCurve.raw_crossings`,
        but without including the closing line between the last
        and first points (i.e. setting include_closure=False).
        '''
        if not virtual_closure:
            return super(OpenKnot, self).raw_crossings(mode=mode,
                                                       include_closure=False,
                                                       recalculate=recalculate,
                                                       try_cython=try_cython)
        cs = super(OpenKnot, self).raw_crossings(mode=mode,
                                                 include_closure=True,
                                                 recalculate=recalculate,
                                                 try_cython=try_cython)

        if len(cs) > 0:
            closure_cs = n.argwhere(((cs[:, 0] > len(self.points)-1)) |
                                    ((cs[:, 1] > len(self.points)-1)))
            indices = closure_cs.flatten()
            for index in indices:
                cs[index, 2] = 0
        return cs
        

    def __str__(self):
        if self._crossings is not None:
            return '<OpenKnot with {} points, {} crossings>'.format(
                len(self.points), len(self._crossings))
        return '<OpenKnot with {} points>'.format(len(self.points))

    def closures(self, quantity, num_closures=10):
        angles = get_rotation_angles(num_closures)

        if not hasattr(Representation, quantity):
            raise ValueError('Representation has no invariant {}'.format(quantity))

        results = []
        print_dist = int(max(1, 3000. / len(self.points)))
        for i, angs in enumerate(angles):
            print('angs are', angs)
            if i % print_dist == 0:
                self._vprint('\ri = {} / {}'.format(i, len(angles)), False)
            k = Knot(self.points, verbose=False)
            k._apply_matrix(rotate_to_top(*angs))
            k.zero_centroid()
            cs = k.raw_crossings()
            if len(cs) > 0:
                closure_cs = n.argwhere(((cs[:, 0] > len(self.points)-1) & (cs[:, 2] < 0.)) |
                                        ((cs[:, 1] > len(self.points)-1) & (cs[:, 2] > 0.)))
                indices = closure_cs.flatten()
                for index in indices:
                    cs[index, 2:] *= -1
            gc = Representation(cs, verbose=False)
            gc.simplify()
            results.append(getattr(gc, quantity)())
        return results

    def arclength(self):
        '''Calls :meth:`pyknotid.spacecurves.spacecurve.SpaceCurve.arclength`,
        automatically *not* including the closure.
        '''
        return super(OpenKnot, self).arclength(include_closure=False)

    def smooth(self, repeats=1, window_len=10, window='hanning'):
        '''Calls :meth:`pyknotid.spacecurves.spacecurve.SpaceCurve.smooth`,
        with the periodic argument automatically set to False.
        '''
        super(OpenKnot, self).smooth(repeats=repeats, window_len=window_len,
                                     window=window, periodic=False)

    def _plot_uniform_angles(self, number_of_samples):
        '''
        Plots the projection of the knot at each of the given
        number of samples, approximately evenly distributed on
        the sphere.

        This function is really just for testing.
        '''
        angles = get_rotation_angles(number_of_samples)

        for i, angs in enumerate(angles):
            k = OpenKnot(self.points, verbose=False)
            k._apply_matrix(rotate_to_top(*angs))
            fig, ax = k.plot_projection(show=False)
            fig.set_size_inches((2, 2))
            fig.savefig('rotation{:05d}'.format(i))
            fig.close()

    def plot_projections(self, number_of_samples):
        '''
        Plots the projection of the knot at each of the given
        number of samples squared, rotated such that the sample
        direction is vertical.

        The output (and return) is a matplotlib plot with
        number_of_samples x number_of_samples axes.
        '''
        angles = get_rotation_angles(number_of_samples ** 2)

        print('Got angles')
        import matplotlib.pyplot as plt

        fig, axes = plt.subplots(nrows=number_of_samples,
                                 ncols=number_of_samples)

        print('got axes')

        all_axes = [ax for row in axes for ax in row]

        for i, angs in enumerate(angles):
            self._vprint('i = {} / {}'.format(i, len(angles)))
            k = OpenKnot(self.points, verbose=False)
            k._apply_matrix(rotate_to_top(*angs))
            ax = all_axes[i]
            fig, ax = k.plot_projection(fig_ax=(fig, ax), show=False)

        fig.tight_layout()
        fig.show()
        return fig, ax

    def alexander_polynomials(self, number_of_samples=10, radius=None,
                              recalculate=False,
                              zero_centroid=False,
                              optimise_closure=True):
        '''
        Returns a list of Alexander polynomials for the knot, closing
        on a sphere of the given radius, with the given number of sample
        points approximately evenly distributed on the sphere.

        The results are cached by number of samples and radius.

        Parameters
        ----------
        number_of_samples : int
            The number of points on the sphere to sample. Defaults to 10.
        optimise_closure: bool
            If True, doesn't really close on a sphere but at infinity.
            This lets the calculation be optimised slightly, and so is the
            default.
        radius : float
            The radius of the sphere on which to close the knot. Defaults
            to None, which picks 10 times the largest Cartesian deviation
            from 0. This is *only* used if optimise_closure=False.
        zero_centroid : bool
            Whether to first move the average position of vertices to
            (0, 0, 0). Defaults to True.

        Returns
        -------
        : ndarray
            A number_of_samples by 3 array of angles and alexander
            polynomials.
        '''
        if zero_centroid:
            self.zero_centroid()

        if not recalculate and self._cached_alexanders is not None:
            if (number_of_samples, radius) in self._cached_alexanders:
                return self._cached_alexanders[(number_of_samples,
                                                radius)]
        else:
            self._cached_alexanders = {}
        angles = get_rotation_angles(number_of_samples)

        polys = []

        cache_radius = radius

        if radius is None:
            radius = 100 * n.max(self.points)
            # Not guaranteed to give 10* the real radius, but good enough

        print_dist = int(max(1, 3000. / len(self.points)))
        try:
            for i, angs in enumerate(angles):
                if i % print_dist == 0:
                    self._vprint('\ri = {} / {}'.format(i, len(angles)), False)
                k = Knot(self.points, verbose=False)
                k._apply_matrix(rotate_to_top(*angs))
                self.k = k 
                if zero_centroid:
                    k.zero_centroid()
                if optimise_closure:
                    cs = k.raw_crossings()
                    if len(cs) > 0:
                        closure_cs = n.argwhere(((cs[:, 0] > len(self.points)-1) & (cs[:, 2] < 0.)) |
                                                ((cs[:, 1] > len(self.points)-1) & (cs[:, 2] > 0.)))
                        indices = closure_cs.flatten()
                        for index in indices:
                            cs[index, 2:] *= -1
                    gc = GaussCode(cs, verbose=self.verbose)
                    gc.simplify()
                    polys.append([angs[0], angs[1], alexander(gc, simplify=False)])

                else:
                    points = k.points
                    closure_point = points[-1] + points[0] / 2.
                    closure_point[2] = radius
                    k.points = n.vstack([points, closure_point])

                    polys.append([angs[0], angs[1], k.alexander_polynomial()])
        except IndexError:
            self.failed_case = k

        self._cached_alexanders[
            (number_of_samples, cache_radius)] = n.array(polys)

        return n.array(polys)

    def closure_alexander_polynomial(self, theta=0, phi=0):
        '''Returns the Alexander polynomial of the knot, when projected in
        the z plane after rotating the given theta and phi to the
        North pole.

        Parameters
        ----------
        theta : float
            The sphere angle theta
        phi : float
            The sphere angle phi
        '''
        k = Knot(self.points, verbose=False)
        k._apply_matrix(rotate_to_top(theta, phi))

        cs = k.raw_crossings()
        if len(cs) > 0:
            closure_cs = n.argwhere(((cs[:, 0] > len(self.points)-1) & (cs[:, 2] < 0.)) |
                                    ((cs[:, 1] > len(self.points)-1) & (cs[:, 2] > 0.)))
            indices = closure_cs.flatten()
            for index in indices:
                cs[index, 2:] *= -1
        gc = GaussCode(cs, verbose=False)
        gc.simplify()
        return alexander(gc, simplify=False)
        

    def alexander_fractions(self, number_of_samples=10, **kwargs):
        '''Returns each of the Alexander polynomials from
        self.alexander_polynomials, with the fraction of that type.
        '''
        polys = self.alexander_polynomials(
            number_of_samples=number_of_samples, **kwargs)
        alexs = n.round(polys[:, 2]).astype(n.int)

        fracs = []
        length = float(len(alexs))
        for alex in n.unique(alexs):
            fracs.append((alex, n.sum(alexs == alex) / length))

        return sorted(fracs, key=lambda j: j[1])

    def _alexander_map_values(self, number_of_samples=10, interpolation=100,
                              **kwargs):
        polys = self.alexander_polynomials(
            number_of_samples=number_of_samples, **kwargs)

        from scipy.interpolate import griddata

        positions = []
        for i, row in enumerate(polys):
            positions.append(gall_peters(row[0], row[1]))
        positions = n.array(positions)

        interpolation_points = n.mgrid[0:2*n.pi:int(1.57*interpolation)*1j,
                                       -2.:2.:interpolation*1j]
        # interpolation_points = n.mgrid[0:2 * n.pi:157j,
        #                        -2.:2.:100j]
        # '''interpolation_points = n.mgrid[-n.pi:n.pi:157j,
        #                        -n.pi/2:n.pi/2:100j]'''

        values = griddata(positions, polys[:, 2],
                          tuple(interpolation_points),
                          method='nearest')

        return positions, values

    def plot_alexander_map(self, number_of_samples=10,
                           scatter_points=False,
                           mode='imshow', interpolation=100,
                           **kwargs):
        '''
        Creates (and returns) a projective diagram showing each
        different Alexander polynomial in a different colour according
        to a closure on a far away point in this direction.

        Parameters
        ----------
        number_of_samples : int
            The number of points on the sphere to close at.
        scatter_points : bool
            If True, plots a dot at each point on the map projection
            where a closure was made.
        mode : str
            'imshow' to plot the pixels of an image, otherwise plots
            filled contours. Defaults to 'imshow'.
        interpolation : int
            The (short) side length of the interpolation grid on which
            the map projection is made. Defaults to 100.
        '''

        positions, values = self._alexander_map_values(
            number_of_samples,
            interpolation=interpolation)

        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        '''fig = plt.figure(figsize=(10, 5))
        ax = fig.add_subplot(111, projection="mollweide")
        ax.grid(True)'''

        if mode == 'imshow':
            cax = ax.imshow(values.T, cmap='jet', interpolation='none')
            fig.colorbar(cax)
        else:
            ax.contourf(values.T, cmap='jet',
                        levels=[0] + range(3, int(n.max(values) + 1.1), 2))
        ax.set_xticks([])
        ax.set_yticks([])

        ax.set_xlim(-0.5, 1.565*interpolation)
        ax.set_ylim(-0.5, 0.995*interpolation)

        im_positions = positions*25*float(interpolation / 100.)
        im_positions[:, 0] -= 0.5
        im_positions[:, 1] += 49.*(interpolation / 100.) + 0.5
        if scatter_points:
            ax.scatter(im_positions[:, 0], im_positions[:, 1], color='black',
                       alpha=1, s=1)

        fig.tight_layout()
        fig.show()
        return fig, ax



    def virtual_check(self):
        '''Takes an open curve and checks (for the default projection) if its
        Gauss code corresponds to a virtual knot or not. Returns a
        Boolean of this information.

        .. warning:: This only checks the distance by which entries in
                     the Gauss code are separated, it is *not*
                     guaranteed to detect virtual knots.

        Returns
        -------
        virtual : bool
            True if the Gauss code corresponds to a virtual knot. False
            otherwise.

        '''
        gauss_code = self.gauss_code()._gauss_code[0][:, 0]
        l = len(gauss_code)
        total_crossings = l / 2
        crossing_counter = 1
        virtual = False

        for crossing_number in self.gauss_code().crossing_numbers:
            occurences = n.where(gauss_code == crossing_number)[0]
            first_occurence = occurences[0]
            second_occurence = occurences[1]
            crossing_difference = second_occurence - first_occurence

            if crossing_difference % 2 == 0:
                return True
        return False


    def virtual_checks(self, number_of_samples=10, 
                                        zero_centroid=False):
        '''
        Returns a list of virtual Booleans for the curve with a given number 
        if projections taken from directions approximately evenly distributed. 
        A value of True corresponds to the projection giving a virtual knot, 
        with False returned otherwise.

        Parameters
        ----------
        number_of_samples : int
            The number of points on the sphere to project from. Defaults to 10.
        zero_centroid : bool
            Whether to first move the average position of vertices to
            (0, 0, 0). Defaults to False.

        Returns
        -------
        : ndarray
            A number_of_samples by 3 array of angles and virtual Booleans
            (True if virtual, False otherwise)
        '''
        if zero_centroid:
            self.zero_centroid()

        angles = get_rotation_angles(number_of_samples)

        polys = []

        print_dist = int(max(1, 3000. / len(self.points)))
        for i, angs in enumerate(angles):
            if i % print_dist == 0:
                self._vprint('\ri = {} / {}'.format(i, len(angles)), False)
            k = OpenKnot(self.points, verbose=False)
            k._apply_matrix(rotate_to_top(*angs))
            if zero_centroid:
                k.zero_centroid()
            isvirtual = k.virtual_check()
            polys.append([angs[0], angs[1], isvirtual])

        return n.array(polys)

    def virtual_fractions(self, number_of_samples=10, **kwargs):
        '''Returns each of the virtual booleans from
        self.virtual.check.projections, with the fraction of each type.
        '''
        polys = self.virtual_checks(
            number_of_samples=number_of_samples, **kwargs)
        alexs = n.round(polys[:, 2]).astype(n.int)

        fracs = []
        length = float(len(alexs))
        for alex in n.unique(alexs):
            fracs.append((alex, n.sum(alexs == alex) / length))

        return sorted(fracs, key=lambda j: j[1])
        
    def _virtual_map_values(self, number_of_samples=10, **kwargs):
        polys = self.virtual_checks(
            number_of_samples=number_of_samples, **kwargs)

        from scipy.interpolate import griddata

        positions = []
        for i, row in enumerate(polys):
            positions.append(gall_peters(row[0], row[1]))
        positions = n.array(positions)

        interpolation_points = n.mgrid[0:2 * n.pi:157j,
                               -2.:2.:100j]
        values = griddata(positions, polys[:, 2],
                          tuple(interpolation_points),
                          method='nearest')

        return positions, values

    def plot_virtual_map(self, number_of_samples=10,
                         scatter_points=False,
                         mode='imshow', **kwargs):
        '''
        Creates (and returns) a projective diagram showing each
        different virtual Boolean in a different colour according
        to a projection in this direction.
        '''

        positions, values = self._virtual_map_values(number_of_samples)

        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()

        if mode == 'imshow':
            cax = ax.imshow(values.T, cmap='jet', interpolation='none')
            fig.colorbar(cax)
        else:
            ax.contourf(values.T, cmap='jet',
                        levels=[0] + range(3, int(n.max(values) + 1.1), 2))
        ax.set_xticks([])
        ax.set_yticks([])

        ax.set_xlim(-0.5, 156.5)
        ax.set_ylim(-0.5, 99.5)

        im_positions = positions * 25
        im_positions[:, 0] -= 0.5
        im_positions[:, 1] += 49.5
        if scatter_points:
            ax.scatter(im_positions[:, 0], im_positions[:, 1], color='black',
                       alpha=1, s=1)

        fig.tight_layout()
        fig.show()

        return fig, ax

    def plot_virtual_shell(self, number_of_samples=10,
                           zero_centroid=False,
                           sphere_radius_factor=2.,
                           opacity=0.3, **kwargs):
        '''
        Plots the curve in 3d via self.plot(), along with a translucent
        sphere coloured according to whether or not the projection from this
        point corresponds to a virtual knot or not.

        Parameters are all passed to 
        :meth:`OpenKnot.virtual_checks`, except opacity and kwargs 
        which are given to mayavi.mesh, and sphere_radius_factor which gives 
        the radius of the enclosing sphere in terms of the maximum Cartesian
        distance of any point in the line from the origin.
        '''

        self.plot()

        positions, values = self._virtual_map_values(
            number_of_samples, zero_centroid=False)

        thetas = n.arcsin(n.linspace(-1, 1, 100)) + n.pi / 2.
        phis = n.linspace(0, 2 * n.pi, 157)

        thetas, phis = n.meshgrid(thetas, phis)

        r = sphere_radius_factor * n.max(self.points)
        zs = r * n.cos(thetas)
        xs = r * n.sin(thetas) * n.cos(phis)
        ys = r * n.sin(thetas) * n.sin(phis)

        import mayavi.mlab as may

        may.mesh(xs, ys, zs, scalars=values, opacity=opacity, **kwargs)

    def self_linking(self, theta=0, phi=0):
        '''
        Takes an open curve, finds its Gauss code (for the default projection)
        and calculates its self linking number, J(K). See Kauffman 2004 for
        more information.

        Returns
        -------
        : self_link_counter : int
            The self linking number of the open curve
        '''
        k = OpenKnot(self.points, verbose=False)
        k._apply_matrix(rotate_to_top(theta, phi))

        # gausscode = k.gauss_code(virtual_closure=True)
        # gausscode.simplify()
        # gausscode = gausscode.without_virtual()._gauss_code[0]
        gausscode = k.gauss_code()._gauss_code[0]
        # gausscode = k.gauss_code()._gauss_code[0]
        l = len(gausscode)

        self_linking_counter = 0

        cache = {}

        for index, row in enumerate(gausscode):
            number, over_under, orientation = row
            if number in cache:
                if ((index - cache[number]) % 2) == 0:
                    self_linking_counter += orientation
            else:
                cache[number] = index

        return self_linking_counter
                    
    def closure_alexander(self, theta=0, phi=0):
        from pyknotid.spacecurves.knot import Knot
        k = Knot(self.points, verbose=False)
        k._apply_matrix(rotate_to_top(theta, phi))

        cs = k.raw_crossings()
        if len(cs) > 0:
            closure_cs = n.argwhere(((cs[:, 0] > len(self.points)-1) & (cs[:, 2] < 0.)) |
                                    ((cs[:, 1] > len(self.points)-1) & (cs[:, 2] > 0.)))
            indices = closure_cs.flatten()
            for index in indices:
                cs[index, 2:] *= -1
        gausscode = GaussCode(cs, verbose=False)

        l = len(gausscode)

        from pyknotid.invariants import alexander
        return alexander(gausscode)
                

    def self_linkings(self, number_of_samples=10,
                      zero_centroid=False, **kwargs):
        '''
        Returns a list of self linking numbers for the curve with a given
        number of projections taken from directions approximately evenly
        distributed.

        Parameters
        ----------
        number_of_samples : int
            The number of points on the sphere to project from. Defaults to 10.
        zero_centroid : bool
            Whether to first move the average position of vertices to
            (0, 0, 0). Defaults to False.

        Returns
        -------
        : ndarray
            A number_of_samples by 3 array of angles and self linking number
        '''
        if zero_centroid:
            self.zero_centroid()

        angles = get_rotation_angles(number_of_samples)

        polys = []

        print_dist = int(max(1, 3000. / len(self.points)))
        for i, angs in enumerate(angles):
            if i % print_dist == 0:
                self._vprint('\ri = {} / {}'.format(i, len(angles)), False)
            k = OpenKnot(self.points, verbose=False)
            k._apply_matrix(rotate_to_top(*angs))
            if zero_centroid:
                k.zero_centroid()
            self_linking = k.self_linking()
            polys.append([angs[0], angs[1], self_linking])

        return n.array(polys)

    def self_linking_fractions(self, number_of_samples=10, **kwargs):
        '''Returns each of the self linking numbers from
        self.virtual.self_link.projections, with the fraction of each type.
        '''
        self_linkings = self.self_linkings(
            number_of_samples=number_of_samples, **kwargs)
        self_linkings = n.round(self_linkings[:, 2]).astype(n.int)

        fracs = []
        length = float(len(self_linkings))
        for alex in n.unique(self_linkings):
            fracs.append((alex, n.sum(self_linkings == alex) / length))
        # fracs = n.array(fracs)

        return sorted(fracs, key=lambda j: j[1])
        #return fracs[n.argsort(fracs[:, 1])]
        
    def _self_linking_map_values(self, number_of_samples=10, **kwargs):
        polys = self.self_linkings(
            number_of_samples=number_of_samples, **kwargs)

        from scipy.interpolate import griddata

        positions = []
        for i, row in enumerate(polys):
            positions.append(gall_peters(row[0], row[1]))
        positions = n.array(positions)

        interpolation_points = n.mgrid[0:2 * n.pi:157j,
                               -2.:2.:100j]
        values = griddata(positions, polys[:, 2],
                          tuple(interpolation_points),
                          method='nearest')

        return positions, values

    def plot_self_linking_map(self, number_of_samples=10,
                           scatter_points=False,
                           mode='imshow', **kwargs):
        '''
        Creates (and returns) a projective diagram showing each
        different self linking number in a different colour according
        to a projection in this direction.
        '''

        positions, values = self._self_linking_map_values(number_of_samples)

        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()

        if mode == 'imshow':
            cax = ax.imshow(values.T, cmap='jet', interpolation='none')
            fig.colorbar(cax)
        else:
            ax.contourf(values.T, cmap='jet',
                        levels=[0] + range(3, int(n.max(values) + 1.1), 2))
        ax.set_xticks([])
        ax.set_yticks([])

        ax.set_xlim(-0.5, 156.5)
        ax.set_ylim(-0.5, 99.5)

        im_positions = positions * 25
        im_positions[:, 0] -= 0.5
        im_positions[:, 1] += 49.5
        if scatter_points:
            ax.scatter(im_positions[:, 0], im_positions[:, 1], color='black',
                       alpha=1, s=1)

        fig.tight_layout()
        fig.show()

        return fig, ax

    def plot_self_linking_shell(self, number_of_samples=100, **kwargs):
        '''
        Plots the curve in 3d via self.plot(), along with a translucent
        sphere coloured by the self linking number obtained by projecting from
        this point.

        Parameters are all passed to 
        :meth:`OpenKnot.virtual_checks`, except opacity and kwargs 
        which are given to mayavi.mesh, and sphere_radius_factor which gives 
        the radius of the enclosing sphere in terms of the maximum Cartesian
        distance of any point in the line from the origin.
        '''
        self.plot(**kwargs)
        plot_shell(self._self_linking_map_values, self.points,
                   number_of_samples=number_of_samples,
                   **kwargs)

    def plot_alexander_shell(self, number_of_samples=100, mode='mesh',
                             radius=None,
                             **kwargs):
        '''
        Plots the curve in 3d via self.plot(), along with a translucent
        sphere coloured by the type of knot obtained by closing on each
        point.

        Parameters are all passed to :meth:`OpenKnot.alexander_polynomials`,
        except opacity and kwargs which are given to mayavi.mesh, and
        sphere_radius_factor which gives the radius of the enclosing
        sphere in terms of the maximum Cartesian distance of any point
        in the line from the origin.
        '''

        self.plot(**kwargs)

        if radius is None:
            self_radii = n.sqrt(n.sum(self.points*self.points, axis=1))
            radius = n.max(self_radii) * 2
        print('radius is', radius)
        if mode == 'mesh':
            plot_sphere_shell_vispy(self.closure_alexander_polynomial,
                                    number_of_samples, number_of_samples,
                                    radius=radius,
                                    **kwargs)
        elif mode == 'crude':
            plot_shell(self._alexander_map_values, self.points,
                       number_of_samples=number_of_samples,
                       **kwargs)


    def alexander_polynomials_multiroots(self, number_of_samples=10,
                                         radius=None,
                                         zero_centroid=False):
        '''
        Returns a list of Alexander polynomials for the knot, closing
        on a sphere of the given radius, with the given number of sample
        points approximately evenly distributed on the sphere. The
        Alexander polynomials are found at three different roots (2, 3
        and 4) and a the knot types corresponding to these roots are
        returned also.

        The results are cached by number of samples and radius.

        Parameters
        ----------
        number_of_samples : int
            The number of points on the sphere to sample. Defaults to 10.
        radius : float
            The radius of the sphere on which to close the knot. Defaults
            to None, which picks 10 times the largest Cartesian deviation
            from 0.
        zero_centroid : bool
            Whether to first move the average position of vertices to
            (0, 0, 0). Defaults to True.

        Returns
        -------
        : ndarray
            A number_of_samples by 3 array of angles and alexander
            polynomials.
        '''
        from pyknotid.catalogue.identify import from_invariants
        from pyknotid.catalogue.database import Knot as dbknot

        if zero_centroid:
            self.zero_centroid()

        if self._cached_alexanders is not None:
            if (number_of_samples, radius) in self._cached_alexanders:
                return self._cached_alexanders[(number_of_samples,
                                                radius)]
        else:
            self._cached_alexanders = {}
        angles = get_rotation_angles(number_of_samples)

        polys = []

        cache_radius = radius

        if radius is None:
            radius = 100 * n.max(self.points)
            # Not guaranteed to give 10* the real radius, but good enough

        print_dist = int(max(1, 3000. / len(self.points)))
        for i, angs in enumerate(angles):
            if i % print_dist == 0:
                self._vprint('\ri = {} / {}'.format(i, len(angles)), False)
            k = Knot(self.points, verbose=False)
            k._apply_matrix(rotate_to_top(*angs))
            if zero_centroid:
                k.zero_centroid()
            points = k.points
            closure_point = points[-1] + points[0] / 2.
            closure_point[2] = radius
            k.points = n.vstack([points, closure_point])
            root_at_two = k.alexander_at_root(2)
            root_at_three = k.alexander_at_root(3)
            root_at_four = k.alexander_at_root(4)
            k_gauss_code = k.gauss_code()
            k_gauss_code.simplify()
            max_crossings = len(k_gauss_code)
            if max_crossings > 17:
                max_crossings = 18
            knot_type = from_invariants(determinant=root_at_two, alex_imag_3=root_at_three,
                                        alex_imag_4=root_at_four,
                                        other=[dbknot.min_crossings <= max_crossings])
            knot_type = knot_db_to_string(knot_type)
            polys.append([angs[0], angs[1], root_at_two, root_at_three, root_at_four, max_crossings,
                          knot_type])

            # self._cached_alexanders[
            #      (number_of_samples, cache_radius)] = n.array(polys)

        return polys

    def multiroots_fractions(self, number_of_samples=10, **kwargs):

        '''
        Returns each of the knot types from
        self.alexander_polynomials_multiroots, with the fraction of that type.
        '''

        knot_info = self.alexander_polynomials_multiroots(
            number_of_samples, **kwargs)
        knot_types = []
        for closures in knot_info:
            knot_types.append(closures[6])
        knot_types = [item for sublist in knot_types for item in sublist] #flattens list
        knot_frequency = Counter(knot_types)
        common = knot_frequency.most_common()
        list_common = [list(elem) for elem in common]
        list_common_fractions = [[elem[0],elem[1]/float(number_of_samples)] for elem in list_common]
        return list_common_fractions

    def generalised_alexander(self):
        '''
        Returns the generalised Alexander polynomial for the default projection
        of the open knot
        '''

        gauss_code_crossings = self.gauss_code()._gauss_code[0][:, 0]
        gauss_code_over_under = self.gauss_code()._gauss_code[0][:,1]
        gauss_code_orientations = self.gauss_code()._gauss_code[0][:,2]
        x = sym.var('x')
        y = sym.var('y')
        m_plus = sym.Matrix([[1 - x, -y], [-x * y**-1, 0]])
        m_minus = sym.Matrix([[0, -x**-1 * y], [-y**-1, 1 - x**-1]])
        num_crossings = len(self.gauss_code())
        matrix = sym.zeros(2 * num_crossings, 2 *num_crossings)
        permutation_matrix = sym.zeros(2 * num_crossings, 2 * num_crossings)

        arc_labels = [0]*len(gauss_code_crossings)
        for i in range(len(gauss_code_crossings)):
            arc_labels[i] = [0] * 4

        counter = 0
        for crossing_number in self.gauss_code().crossing_numbers:
            occurrences = n.where(gauss_code_crossings == crossing_number)[0]
            if gauss_code_orientations[occurrences[0]] == 1:
                m = m_plus
            else:
                m = m_minus
            for i in [0,1]:
                for j in [0,1]:
                    matrix[counter*2 + i,counter*2 + j] = m[i,j]
            counter += 1

        for i in range(len(gauss_code_crossings)):
            arc_labels[i][0] = gauss_code_crossings[i]
            arc_labels[i][1] = (gauss_code_orientations[i] *
                                gauss_code_over_under[i]) #-1 = r, +1 = l
            arc_labels[i-1][2] = gauss_code_crossings[i]
            arc_labels[i-1][3] = (-1 * gauss_code_orientations[i] *
                                  gauss_code_over_under[i]) #-1 = r, +1 = l

        counter = 1
        for crossing_number in self.gauss_code().crossing_numbers:
            for i in range(len(gauss_code_crossings)):
                if arc_labels[i][0] == crossing_number:
                    arc_labels[i][0] = counter
                if arc_labels[i][2] == crossing_number:
                    arc_labels[i][2] = counter
            counter += 1

        for i in range(len(arc_labels)):
            if arc_labels[i][1] < 0:
                if arc_labels[i][3] < 0 :
                    #bottom right
                    permutation_matrix[arc_labels[i][2]*2-1, arc_labels[i][0]*2-1] = 1
                else:
                    #bottom left
                    permutation_matrix[arc_labels[i][2]*2-2, arc_labels[i][0]*2-1] = 1
            else:
                if arc_labels[i][3] < 0:
                    #upper right
                    permutation_matrix[arc_labels[i][2]*2-1, arc_labels[i][0]*2-2] = 1
                else:
                    #upper left
                    permutation_matrix[arc_labels[i][2]*2-2, arc_labels[i][0]*2-2] = 1

        writhe = sum(gauss_code_orientations)/2

        return (-1)**writhe * ((matrix - permutation_matrix).det())

    def projection_invariant(self, **kwargs):
        '''
        First checks if the projection of an open curve is virtual or classical. If virtual,
        a virtual knot invariant is calculated. Otherwise a classical invariant is calculated.
        '''
        is_virtual = self.virtual_check()
        if is_virtual:
            return(['v', self.self_linking()])
        else:
            from pyknotid.invariants import alexander
            return(['c'], alexander(self.gauss_code(), variable=-1,
                                    quadrant='lr',
                                    mode='python', simplify=False))

    def slip_vassiliev_degree_2_average(self, samples=10, recalculate=False, **kwargs):
        from pyknotid.spacecurves.rotation import (get_rotation_angles,
                                                  rotate_to_top)
        from pyknotid.spacecurves import Knot
        angles = get_rotation_angles(samples)
    
        v2s = []
        for theta, phi in angles:
            k = OpenKnot(self.points, verbose=False)
            k._apply_matrix(rotate_to_top(theta, phi))
            v2 = k._slip_vassiliev_degree_2_projection()
            v2s.append(v2)

        result = n.average(n.abs(v2s))
        return result

    def _slip_vassiliev_degree_2_projection(self):
        gc = self.gauss_code()
        from pyknotid.writhes import slip_vassiliev_2
        return slip_vassiliev_2(gc)

    def vassiliev_degree_2_average(self, samples=10, recalculate=False, **kwargs):
        '''Returns the average Vassliev degree 2 invariant calculated by
        averaging its combinatorial value over many different
        projection directions.

        Parameters
        ----------
        samples : int
            The number of directions to average over. Defaults to 10.
        recalculate : bool
            Whether to recalculate the writhe even if a cached result
            is available. Defaults to False.
        **kwargs :
            These are passed directly to :meth:`raw_crossings`.
        '''
        if (self._cached_v2 and
            samples in self._cached_v2 and
            not recalculate):
            return self._cached_v2[samples]

        from pyknotid.spacecurves.rotation import (get_rotation_angles,
                                                  rotate_to_top)
        from pyknotid.spacecurves import Knot
        angles = get_rotation_angles(samples)
    
        v2s = []
        for theta, phi in angles:
            k = OpenKnot(self.points, verbose=False)
            k._apply_matrix(rotate_to_top(theta, phi))
            v2 = k._vassiliev_degree_2_projection()
            v2s.append(v2)

        result = n.average(n.abs(v2s))
        self._cached_v2[samples] = result
        return result

    def _vassiliev_degree_2_projection(self):
        from pyknotid.spacecurves import Knot
        k = Knot(self.points, verbose=False)
        return k.vassiliev_degree_2(simplify=False,
                                    include_closure=False)

    def vassiliev_degree_3_average(self, samples=10, recalculate=False,
                                   signed=True, **kwargs):
        if (self._cached_v3 and
            samples in self._cached_v3 and
            not recalculate):
            return self._cached_v3[samples]

        from pyknotid.spacecurves.rotation import get_rotation_angles, rotate_to_top            
        from pyknotid.spacecurves import Knot
        angles = get_rotation_angles(samples)
    
        v3s = []
        for theta, phi in angles:
            k = OpenKnot(self.points, verbose=False)
            k._apply_matrix(rotate_to_top(theta, phi))
            v3 = k._vassiliev_degree_3_projection()
            v3s.append(v3)

        if signed:
            result = n.average(v3s)
        elif signed is None:
            result = (n.average(v3s), n.average(n.abs(v3s)))
        else:
            result = n.average(n.abs(v3s))
        self._cached_v3[(samples, signed)] = result
        return result

    def _vassiliev_degree_3_projection(self):
        from pyknotid.spacecurves import Knot
        k = Knot(self.points, verbose=False)
        return k.vassiliev_degree_3(simplify=False,
                                    include_closure=False)

    def virtual_vassiliev_degree_3(self):
        from pyknotid.invariants import virtual_vassiliev_degree_3
        return virtual_vassiliev_degree_3(self.gauss_code())

    def _determinants_and_self_linkings(self, number_of_samples=10, radius=None,
                                        recalculate=False,
                                        zero_centroid=False):
        if zero_centroid:
            self.zero_centroid()

        angles = get_rotation_angles(number_of_samples)

        polys = []
        self_linkings = []

        cache_radius = radius

        if radius is None:
            radius = 100 * n.max(self.points)
            # Not guaranteed to give 10* the real radius, but good enough

        print_dist = int(max(1, 3000. / len(self.points)))
        for i, angs in enumerate(angles):
            if i % print_dist == 0:
                self._vprint('\ri = {} / {}'.format(i, len(angles)), False)
            k = Knot(self.points, verbose=False)
            k._apply_matrix(rotate_to_top(*angs))
            if zero_centroid:
                k.zero_centroid()
            cs = k.raw_crossings()
            if len(cs) > 0:
                closure_cs = n.argwhere(((cs[:, 0] > len(self.points)-1) & (cs[:, 2] < 0.)) |
                                        ((cs[:, 1] > len(self.points)-1) & (cs[:, 2] > 0.)))
                indices = closure_cs.flatten()
                for index in indices:
                    cs[index, 2:] *= -1
            gc = GaussCode(cs, verbose=self.verbose)
            gc.simplify()
            polys.append([angs[0], angs[1], alexander(gc, simplify=False)])

            # Remove closing crossings to calculate self linking
            xs, ys = n.where(cs[:, :2] > len(k.points) - 1)
            keeps = n.ones(len(cs), dtype=n.bool)
            for x in xs:
                keeps[x] = False
            cs = cs[keeps]
            gausscode = GaussCode(cs)._gauss_code[0]
            l = len(gausscode)
            self_linking_counter = 0
            cache = {}
            for index, row in enumerate(gausscode):
                number, over_under, orientation = row
                if number in cache:
                    if ((index - cache[number]) % 2) == 0:
                        self_linking_counter += orientation
                else:
                    cache[number] = index
            self_linkings.append((angs[0], angs[1], self_linking_counter))


        return n.array(polys), n.array(self_linkings)

    def _determinant_and_self_linking_fractions(self, number_of_samples=10,
                                                **kwargs):
        polys, self_linkings = self._determinants_and_self_linkings(
            number_of_samples, **kwargs)

        alexs = n.round(polys[:, 2]).astype(n.int)

        fracs = []
        length = float(len(alexs))
        for alex in n.unique(alexs):
            fracs.append((alex, n.sum(alexs == alex) / length))

        det_fracs = sorted(fracs, key=lambda j: j[1])


        self_linkings = n.round(self_linkings[:, 2]).astype(n.int)

        fracs = []
        length = float(len(self_linkings))
        for linking in n.unique(self_linkings):
            fracs.append((linking, n.sum(self_linkings == linking) / length))

        self_linking_fracs = sorted(fracs, key=lambda j: j[1])

        return det_fracs, self_linking_fracs


    def _closure_and_projection_invariants(self, number_of_samples=10, radius=None,
                                           recalculate=False,
                                           zero_centroid=False):
        if zero_centroid:
            self.zero_centroid()

        angles = get_rotation_angles(number_of_samples)

        closure_knotted = []
        projection_virtual = []
        projection_knotted = []

        cache_radius = radius

        if radius is None:
            radius = 100 * n.max(self.points)
            # Not guaranteed to give 10* the real radius, but good enough

        print_dist = int(max(1, 3000. / len(self.points)))
        for i, angs in enumerate(angles):
            if i % print_dist == 0:
                self._vprint('\ri = {} / {}'.format(i, len(angles)), False)
            k = Knot(self.points, verbose=False)
            k._apply_matrix(rotate_to_top(*angs))
            if zero_centroid:
                k.zero_centroid()
            cs = k.raw_crossings()
            if len(cs) > 0:
                closure_cs = n.argwhere(((cs[:, 0] > len(self.points)-1) & (cs[:, 2] < 0.)) |
                                        ((cs[:, 1] > len(self.points)-1) & (cs[:, 2] > 0.)))
                indices = closure_cs.flatten()
                for index in indices:
                    cs[index, 2:] *= -1
            gc = GaussCode(cs, verbose=self.verbose)
            gc.simplify()
            is_knotted = (alexander(gc, simplify=False)) > 1.5
            closure_knotted.append([angs[0], angs[1], is_knotted])

            # Remove closing crossings to calculate self linking
            xs, ys = n.where(cs[:, :2] > len(k.points) - 1)
            keeps = n.ones(len(cs), dtype=n.bool)
            for x in xs:
                keeps[x] = False
            cs = cs[keeps]
            rep = Representation(cs)
            is_virtual = rep.self_linking() != 0
            projection_virtual.append((angs[0], angs[1], is_virtual))

            if not is_virtual:
                is_knotted = alexander(gc, simplify=False) > 1.5
                projection_knotted.append([angs[0], angs[1], is_knotted])
            else:
                projection_knotted.append([angs[0], angs[1], False])


        return (n.array(closure_knotted), n.array(projection_virtual),
                n.array(projection_knotted))

    def _closure_and_projection_knotted_fractions(self, number_of_samples=10,
                                                    **kwargs):
        polys, self_linkings = self._determinants_and_self_linkings(
            number_of_samples, **kwargs)
        ck, pv, pk = self._closure_and_projection_invariants(number_of_samples, **kwargs)

        ck_fraction = n.average(ck[:, -1])
        pv_fraction = n.average(pv[:, -1])

        return ck_fraction, pv_fraction, n.average(pv[:, -1].astype(n.bool) |
                                                   pk[:, -1].astype(n.bool))

    def plot_vassiliev_spectrum(self, angles=100, max_crossings=8):
        v2 = self.vassiliev_degree_2_average()
        v3 = self.vassiliev_degree_3_average()

        from pyknotid.catalogue.identify import from_invariants
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()

        ks = from_invariants(max_crossings=max_crossings)
        for k in ks:
            kv2 = k.vassiliev_2
            kv3 = k.vassiliev_3

            ax.scatter([kv2], [kv3])

        ax.set_xlabel('v2')
        ax.set_ylabel('v3')

        ax.scatter([v2], [v3], color='red')

        fig.show()

        return fig, ax


def gall_peters(theta, phi):
    '''
    Converts spherical coordinates to the Gall-Peters
    projection of the sphere, an area-preserving projection in
    the shape of a Rectangle.

    Parameters
    ----------
    theta : float
        The latitude, in radians.
    phi : float
        The longitude, in radians.
    '''
    theta -= n.pi / 2
    return (phi, 2 * n.sin(theta))

def mollweide(phi, lambda_):
    '''

    Converts spherical coordinates to the Mollweide
    projection of the sphere, an area-preserving projection in
    the shape of an ellipse.

    Parameters
    ----------
    phi : float
        The latitude, in radians.
    lambda_ : float
        The longitude, in radians.
    '''

    if phi == n.pi/2 or phi == -n.pi/2:
        theta_m = phi
    else:
        theta_n = phi
        theta_m = 0
        while abs(theta_m - theta_n) > 0.000001:
            theta_n = theta_m
            theta_m = (theta_n - (2*theta_n + n.sin(2*theta_n) - n.pi * n.sin(phi)) /
                       (2 + 2*n.cos(2*theta_n)))
    x = ((2 * n.sqrt(2)) / (n.pi)) * lambda_ * n.cos(theta_m)
    y = n.sqrt(2) * n.sin(theta_m)
    return(x,y)


def knot_db_to_string(database_object):
    '''
    Takes output from from_invariants() and returns knot type as decimal.
    For example: <Knot 3_1> becomes 3.1 and <Knot K13n1496> becomes 13.1496
    '''
    db_strings = []
    for entries in database_object:
        db_string = str(entries)
        if db_string[6] == 'K':
            db_string = db_string[7:-1]
        else:
            db_string = db_string[6:-1]
            db_string = db_string.replace('_', '.')
        db_strings.append(db_string)
    return db_strings


